"""
This module contains various functions to create plots using MESA and/or GYRE data.

Examples:
    Create an HRD from a history file from the MS onwards using a simple mask to skip the PMS.

    >>> import wsssss.load_data as ld
    >>> import wsssss.plotting.plotting as pl
    >>> hist = ld.History('path/to/LOGS/history.data')
    >>> pl.make_hrd(hist, use_mask=hist.data.center_h1<hist.data.center_h1[0] * 0.99)

"""

import copy
import os

import matplotlib as mpl
import numpy as np
from matplotlib import animation
from matplotlib import colors
from matplotlib import pyplot as plt
from matplotlib import transforms
from matplotlib import cm
from matplotlib.collections import LineCollection
from matplotlib.lines import Line2D
from scipy.interpolate import interp1d

from . import utils as pu
from .. import functions as uf
from . import kipper

np.seterr(invalid='ignore')


def calc_vhrd(hist, norm):
    xdat, ydat = uf.get_logTeffL(hist)

    tdat = np.diff(hist.get('star_age'), prepend=0.0)
    # tdat = 10 ** hist.get('log_dt')
    dHRD = np.sqrt(np.diff(np.append(0, xdat)) ** 2 + np.diff(np.append(0, ydat)) ** 2)
    vHRD = dHRD / tdat / norm
    return vHRD


def plot_vhrd(f, ax, hist, vHRD_norm, use_mask, norm, add_cbar, add_label):
    ini_mass = hist.get('star_mass')[0]
    xdat, ydat = uf.get_logTeffL(hist)

    mask = uf.get_mask(hist, use_mask)

    xdat = xdat[mask]
    ydat = ydat[mask]
    vHRD = calc_vhrd(hist, vHRD_norm)[mask]

    cax = ax.scatter(xdat, ydat, c=np.log10(vHRD), lw=0, cmap='viridis', norm=norm)
    ax.plot(xdat, ydat, c='k', lw=1)
    cbar = None
    if add_cbar:
        cbar = f.colorbar(cax, ax=ax)

    if add_label:
        ax.text(xdat[0], ydat[0], '${:.2f}\\mathrm{{M}}_\\odot$'.format(ini_mass), ha='right', va='top', fontsize=10)

    return f, ax, cbar, cax


def make_vhrd(hist, add_cbar=True, vHRD_norm=1.0, use_mask=True, ax=None, bounds=None, add_label=False):
    if bounds is None:
        bounds = np.arange(-9, 0, 1)
    norm = colors.BoundaryNorm(boundaries=bounds, ncolors=256)

    f, ax = pu.get_figure(ax)

    if type(hist) in [list, tuple, dict]:
        for h in hist:
            if isinstance(hist, dict):
                h = hist[h]
            f, ax, cbar, cax = plot_vhrd(f, ax, h, vHRD_norm, use_mask, norm, add_cbar, add_label)
            add_cbar = False
    else:
        f, ax, cbar, cax = plot_vhrd(f, ax, hist, vHRD_norm, use_mask, norm, add_cbar, add_label)

    if add_cbar:
        cbax = f.axes[1]
        cbax.set_ylabel(r'$\log_{10} \; v_\mathrm{HRD}$')

    ax.set_xlabel(r'$\log_{10} \; (T_\mathrm{eff} / K)$')
    ax.set_ylabel(r'$\log_{10} \; (L / L_\odot)$')

    ax.invert_xaxis()

    pu.hrd_const_rad(ax, fontsize=10)

    return f, ax


def plot_hrd_data(f, ax, hist, zdata, znorm, use_mask, norm, add_cbar, cmap, sdat=None, small_cbar=False,
                  label_func=None, vmin=None, vmax=None, linecolor='k', linear=False, linestyle='-'):
    xdat, ydat = uf.get_logTeffL(hist)
    if linear:
        xdat = 10 ** xdat
        ydat = 10 ** ydat

    mask = uf.get_mask(hist, use_mask)
    if not np.any(mask):
        return f, ax, None, None

    if type(zdata) == str:
        zdat = hist.get(zdata) / znorm
    elif zdata is None:
        zdat = np.zeros_like(xdat)
    elif np.ndim(zdata) == 0:
        zdat = (zdata / znorm) * np.ones_like(xdat)
    else:
        zdat = zdata / znorm

    xdat = xdat[mask]
    ydat = ydat[mask]
    zdat = zdat[mask]
    if sdat is not None:
        sdat = sdat[mask][::-1]
    else:
        sdat = 36

    # Reverse data to plot early data on top.
    if zdata is not None:
        ax.scatter(xdat[::-1], ydat[::-1], c='k', s=7 ** 2)
        cax = ax.scatter(xdat[::-1], ydat[::-1], c=zdat[::-1], s=sdat, lw=0, cmap=cmap, norm=norm, vmin=vmin, vmax=vmax)
    else:
        cax = None
    ax.plot(xdat, ydat, c=linecolor, lw=1, linestyle=linestyle)
    cbar = None
    if add_cbar and zdata is not None:
        if not small_cbar:
            cbar = f.colorbar(cax, ax=ax)
        else:
            cbar = f.colorbar(cax, ax=ax, aspect=40, pad=0)

    if label_func is None:
        ini_mass = hist.get('star_mass')[0]
        label = '${:.1f}M_\\odot$'.format(ini_mass)
        i_label = 0
        ha = 'right'
        va = 'top'
        angle = 0
    else:
        label, i_label, ha, va, angle = label_func(hist, mask)

    ax.text(xdat[i_label], ydat[i_label], label, ha=ha, va=va, fontsize=10, rotation=angle)

    return f, ax, cbar, cax


def make_hrd(hist, zdata=None, zlabel='', add_cbar=True, znorm=1.0, use_mask=None, ax=None, norm=None, cmap='viridis',
             add_const_rad=True, label_func=None, vmin=None, vmax=None, linecolor='k', linear=False):
    f, ax = pu.get_figure(ax)
    linestyles = ['-', '--', '-.', ':']
    if type(hist) in [list, tuple]:
        pass
    else:
        hist = [hist]
    if (zdata is not None) and not callable(zdata):
        if len(hist) == 1:
            if len(zdata) == len(hist[0]):
                zdata = [zdata]
        if (len(zdata) != len(hist)) and (type(zdata) != str):
            raise ValueError("When passing multiple histories and zdata is not a string"
                             "the length of zdata must the the length of hist.")

    for i, h in enumerate(hist):
        if type(zdata) == str or zdata is None:
            f, ax, cbar, cax = plot_hrd_data(f, ax, h, zdata, znorm, use_mask, norm, add_cbar, cmap, vmin=vmin,
                                             vmax=vmax, label_func=label_func, linecolor=linecolor, linear=linear,
                                             linestyle=linestyles[i % len(linestyles)])
        elif callable(zdata):
            f, ax, cbar, cax = plot_hrd_data(f, ax, h, zdata(h), znorm, use_mask, norm, add_cbar, cmap, vmin=vmin,
                                             vmax=vmax, label_func=label_func, linecolor=linecolor, linear=linear,
                                             linestyle=linestyles[i % len(linestyles)])
        else:
            f, ax, cbar, cax = plot_hrd_data(f, ax, h, zdata[i], znorm, use_mask, norm, add_cbar, cmap, vmin=vmin,
                                             vmax=vmax, label_func=label_func, linecolor=linecolor, linear=linear,
                                             linestyle=linestyles[i % len(linestyles)])
        if add_cbar and cbar is not None:
            cbax = f.axes[-1]
            cbax.set_ylabel('{}'.format(zlabel))
        add_cbar = False

    if linear:
        ax.set_xlabel(r'$T_\mathrm{eff} \; (\mathrm{K})$')
        ax.set_ylabel(r'$L \; (\mathrm{L}_\odot)$')
    else:
        ax.set_xlabel(r'$\log_{10} \; (T_\mathrm{eff} / \mathrm{K})$')
        ax.set_ylabel(r'$\log_{10} \; (L / \mathrm{L}_\odot)$')

    if np.diff(ax.get_xlim()) > 0:
        ax.invert_xaxis()
    ax.margins(0.05)

    if add_const_rad:
        pu.hrd_const_rad(ax, linear=linear)

    return f, ax


def make_propagation(p, hist, xname='logR', l=1, ax=None, only_NS=True, do_reduced=False, only_reduced=False,
                     add_legend=True, n_col_legend=3, fill_cavity=False, show_burn_level=0, do_numax=True):
    c = uf.get_constants(p)

    f, ax = pu.get_figure(ax)
    radius = uf.get_radius(p)

    csound = p.data.csound
    brunt_N2 = p.data.brunt_N2
    brunt_N2[brunt_N2 < 0] = 0

    lamb2 = uf.get_lamb2(p)

    if hist is not None:
        i_hist = p.get_hist_index(hist)

    if 'scale_height' in p.columns:
        H = p.data.scale_height * c.rsun
        cs2H2 = (csound / (4 * np.pi * H)) ** 2
        nu_c2 = cs2H2[1:] * (1 - 2 * np.diff(H) / np.diff(radius))
        x_skip = 1
    else:
        rho = 10 ** p.data.logRho
        dlnrho = -np.diff(np.log(rho))
        H = -(dlnrho / -np.diff(radius)) ** -1
        cs2H2 = (csound[1:] / (4 * np.pi * H)) ** 2
        nu_c2 = cs2H2[1:] * (1 - 2 * np.diff(H) / np.diff(radius)[1:])
        x_skip = 2

    x = pu.get_x_and_set_xlabel(p, xname, ax, hist=hist)

    if not only_reduced:  # Also plot normal N and S
        S = 1E6 / (2 * np.pi) * lamb2 ** 0.5
        N = 1E6 / (2 * np.pi) * brunt_N2 ** 0.5
        ax.plot(x, S, 'C0', label=r'$S_{\ell=1}$')
        ax.plot(x, N, 'C1', label='$N$')

    if do_reduced:  # Add N_red and S_red
        if 'q_J' in p.columns:
            J = p.get('q_J')
        else:
            J = 1 - 1 / 3 * np.gradient(np.log(p.data.mass), np.log(radius))

        S_red = 1E6 / (2 * np.pi) * lamb2 ** 0.5 * J
        N_red = 1E6 / (2 * np.pi) * brunt_N2 ** 0.5 / J

        ax.plot(x, S_red, 'C0--', label='$\\tilde{S}_{\\ell=1}$')
        ax.plot(x, N_red, 'C1--', label='$\\tilde{N}$')

    if not only_NS and not only_reduced:  # Add Acoustic cutouff
        ax.plot(x[x_skip:], 1E6 * nu_c2 ** 0.5, label=r'$\nu_c$')
        ax.plot(x, 1E6 * (csound / (4 * np.pi * H)), ls='--', label=r'$\nu_\mathrm{ac}$ approx')
        if hist is not None:
            ax.hlines(hist.data.acoustic_cutoff[i_hist] / (2 * np.pi), x[0], x[-1], label=r'$\nu_\mathrm{ac}$')

    if do_numax and (hist is not None):
        numax = hist.get('nu_max')[i_hist]
        xmin = np.nanmin(x)
        xmax = np.nanmax(x)

        if do_reduced:
            y1 = S_red
            y2 = N_red
        else:
            y1 = S
            y2 = N

        gmask = numax <= np.minimum(y1, y2)  # g-like
        pmask = numax >= np.maximum(y1, y2)  # p-like
        emask = ~(pmask | gmask)  # not p-like or g-like

        nu1, = ax.plot(x, np.where(~gmask, np.nan, numax * gmask), 'k--')
        nu2, = ax.plot(x, np.where(~emask, np.nan, numax * emask), 'k:')
        nu3, = ax.plot(x, np.where(~pmask, np.nan, numax * pmask), 'k-')

        nu_lines = (nu1, nu2, nu3)
        nu_lines_lc = mpl.collections.LineCollection([[[0, 0]] * 2] * 3, linestyles=['--', ':', '-'],
                                                     colors=['k', 'k', 'k'], label=r'$\nu_\mathrm{max}$')

    if xname != 's':
        ax.set_ylim(1e6 * np.min(lamb2) ** 0.5 / 2, 1e6 * np.max(lamb2) ** 0.5 * 2)
    ax.set_ylabel(r'Frequency ($\mu$Hz)')

    ax.set_yscale('log')
    if xname in ['logR']:
        ax.set_xscale('log')

    if fill_cavity:
        if do_reduced:
            ax.fill_between(x, -1e99 * np.ones_like(S_red), np.minimum(S_red, N_red), label='g-cavity', hatch='//',
                            zorder=-5,
                            facecolor='none', edgecolor='C1')
            ax.fill_between(x, np.maximum(S_red, N_red), 1e99, label='p-cavity', hatch='--', zorder=-5,
                            facecolor='none', edgecolor='C0')
        else:
            ax.fill_between(x, -1e99 * np.ones_like(S), np.minimum(S, N), label='g-cavity', hatch='//',
                            zorder=-5,
                            facecolor='none', edgecolor='C1')
            ax.fill_between(x, np.maximum(S, N), 1e99, label='p-cavity', hatch='--', zorder=-5,
                            facecolor='none', edgecolor='C0')

    if show_burn_level > 0:
        burning = p.data.net_nuclear_energy >= show_burn_level

        h_mask = (np.gradient(p.data.h1, p.data.mass) > 0) & (p.data.h1 > 1e-6) & burning
        he_mask = (np.gradient(p.data.he4, p.data.mass) > 0) & (p.data.he4 > 1e-6) & (p.data.h1 < 1e-6) & burning

        if np.any(h_mask):
            bmin_h = min(x[h_mask])
            bmax_h = max(x[h_mask])
            ax.axvspan(bmin_h, bmax_h, color='tab:cyan', zorder=-4, label='H-burn', alpha=0.5)

        if np.any(he_mask):
            bmin_he = min(x[he_mask])
            bmax_he = max(x[he_mask])
            ax.axvspan(bmin_he, bmax_he, color='tab:olive', zorder=-4, label='He-burn', alpha=0.5)

    handles, labels = ax.get_legend_handles_labels()
    handler_map = None
    if do_numax and (hist is not None):
        labels.append(nu_lines_lc.get_label())
        handles.append(nu_lines_lc)
        handler_map = {LineCollection: pu.HandlerDashedLines()}

    if add_legend:
        n_item = len(labels)
        n_row = int(np.ceil(n_item / n_col_legend))

        order = list(range(n_item))
        if 'p-cavity' in labels and 'He-burn' in labels:
            order[labels.index('He-burn')] = labels.index('p-cavity')
            order[labels.index('p-cavity')] = labels.index('He-burn')
        elif 'p-cavity' in labels:
            order[labels.index('p-cavity')] = labels.index('g-cavity')
            order[labels.index('g-cavity')] = labels.index('p-cavity')

        labels = [f'{labels[i]}' for i in range(n_item)]
        ax.legend([handles[idx] for idx in order], [labels[idx] for idx in order], ncol=n_col_legend,
                  bbox_to_anchor=(0., 1.02, 1., .102), loc=3, mode='expand', borderaxespad=0.,
                  handler_map=handler_map)
    return f, ax


def make_propagation2(p, hist, xname='logR', l=1, ax=None, only_NS=True, do_reduced=True, only_reduced=False,
                      add_legend=True, fill_cavity=False, show_burn_level=0):
    c = uf.get_constants(p)

    f, ax = pu.get_figure(ax)

    hist_i = hist.get_profile_index(p.profile_num)[0]

    radius = uf.get_radius(p)

    csound = p.data.csound
    brunt_N2 = p.data.brunt_N2
    # brunt_N2[brunt_N2 < 0] = 0

    lamb2 = uf.get_lamb2(p)

    if 'scale_height' in p.columns:
        H = p.data.scale_height * c.rsun
        cs2H2 = (csound / (4 * np.pi * H)) ** 2
        nu_c2 = cs2H2[1:] * (1 - 2 * np.diff(H) / np.diff(radius))
        x_skip = 1
    else:
        rho = 10 ** p.data.logRho
        dlnrho = -np.diff(np.log(rho))
        H = -(dlnrho / -np.diff(radius)) ** -1
        cs2H2 = (csound[1:] / (4 * np.pi * H)) ** 2
        nu_c2 = cs2H2[1:] * (1 - 2 * np.diff(H) / np.diff(radius)[1:])
        x_skip = 2

    if xname == 's':
        if hist is None:
            raise ValueError("Need History for xname='s'.")
        else:
            i_hist = p.get_hist_index(hist)
            r1 = hist.data.r_1[i_hist]
            r2 = hist.data.r_2[i_hist]
        r0 = np.sqrt(r1 * r2)
        x = np.log(radius / r0)
        s0 = np.log(r1 / r2) / 2
        ax.set_xlabel('$s$')
        ax.set_xlim(-1.5 * abs(s0), 1.5 * abs(s0))
        smin, smax = ax.get_xlim()
        x[(x < smin) | (x > smax)] = np.nan
    else:
        x = pu.get_x_and_set_xlabel(p, xname, ax, hist=hist)

    if not only_reduced:  # Also plot normal N and S
        S2 = (1E6 / (2 * np.pi)) ** 2 * lamb2
        N2 = (1E6 / (2 * np.pi)) ** 2 * brunt_N2
        ax.plot(x, S2, label=r'$S^2_{\ell=1}$')
        ax.plot(x, N2, label='$N^2$')

    ax.axhline(hist.get('nu_max')[hist_i] ** 2, c='k', zorder=-1, label=r'$\nu^2_\mathrm{max}$')

    if do_reduced:  # Add N_red and S_red
        J = 1 - 1 / 3 * np.gradient(np.log(p.data.mass), np.log(radius))
        S2_red = (1E6 / (2 * np.pi) * J) ** 2 * lamb2
        N2_red = (1E6 / (2 * np.pi) / J) ** 2 * brunt_N2

        ax.plot(x, S2_red, label='$\\tilde{S}^2_{\\ell=1}$')
        ax.plot(x, N2_red, label='$\\tilde{N}^2$')

    if not only_NS and not only_reduced:  # Add Acoustic cutouff
        ax.plot(x[x_skip:], 1e12 * nu_c2, label=r'$\nu^2_\mathrm{ac}$')
        ax.plot(x, 1e12 * (csound / (4 * np.pi * H)) ** 2, ls='--', label=r'$\nu^2_\mathrm{ac}$ approx')
        ax.hlines((hist.data.acoustic_cutoff[hist_i] / (2 * np.pi)) ** 2, x[0], x[-1], label=r'$\nu^2_\mathrm{ac}$')

    ax.set_ylim(1e12 * np.min(lamb2) / 2, 1e12 * np.max(lamb2) * 2)
    ax.set_ylabel(r'Frequency$^2$ ($\mu$Hz$^2$)')

    ax.set_yscale('log')
    if xname in ['logR']:
        ax.set_xscale('log')

    if fill_cavity:
        if do_reduced:
            ax.fill_between(x, -1e99 * np.ones_like(S2_red), np.minimum(S2_red, N2_red), label='g-cavity', hatch='//',
                            zorder=-5,
                            facecolor='none', edgecolor='C5')
            ax.fill_between(x, np.maximum(S2_red, N2_red), 1e99, label='p-cavity', hatch='--', zorder=-5,
                            facecolor='none', edgecolor='C8')

    if show_burn_level > 0:
        burning = p.data.net_nuclear_energy >= show_burn_level

        h_mask = (np.gradient(p.data.h1, p.data.mass) > 0) & (p.data.h1 > 1e-6) & burning
        he_mask = (np.gradient(p.data.he4, p.data.mass) > 0) & (p.data.he4 > 1e-6) & (p.data.h1 < 1e-6) & burning

        if np.any(h_mask):
            bmin_h = min(x[h_mask])
            bmax_h = max(x[h_mask])
            ax.axvspan(bmin_h, bmax_h, color='tab:cyan', zorder=-4, label='H-burn', alpha=0.5)

        if np.any(he_mask):
            bmin_he = min(x[he_mask])
            bmax_he = max(x[he_mask])
            ax.axvspan(bmin_he, bmax_he, color='tab:olive', zorder=-4, label='He-burn', alpha=0.5)

    if add_legend:
        handles, labels = ax.get_legend_handles_labels()
        n_item = len(labels)
        n_row = int(np.ceil(n_item / 3))
        n_col = 3

        order = list(range(n_item))
        if 'p-cavity' in labels and 'He-burn' in labels:
            order[labels.index('He-burn')] = labels.index('p-cavity')
            order[labels.index('p-cavity')] = labels.index('He-burn')
        elif 'p-cavity' in labels:
            order[labels.index('p-cavity')] = labels.index('g-cavity')
            order[labels.index('g-cavity')] = labels.index('p-cavity')

        labels = [f'{labels[i]}' for i in range(n_item)]
        ax.legend([handles[idx] for idx in order], [labels[idx] for idx in order], ncol=n_col,
                  bbox_to_anchor=(0., 1.02, 1., .102), loc=3, mode='expand', borderaxespad=0.)
    return f, ax


def make_echelle(gs, hist=None, ax=None, l_list=(0, 1, 2), offset='auto', delta_nu='weighted',
                 prefix='profile', suffix='.data.GYRE.sgyre_l', freq_units='uHz', legend_loc='upper right'):
    f, ax = pu.get_figure(ax)

    if hist is not None:
        pnum = int(gs.path.split(prefix)[-1].replace(suffix, ''))
        hist_i = hist.get_profile_index(pnum)[0]
    nu_all = gs.get_frequencies(freq_units)

    if type(delta_nu) == str:
        delta_nu = delta_nu.lower()

        if delta_nu == 'mesa':
            delta_nu = hist.data.delta_nu[hist_i]

        elif delta_nu in ['mean', 'median', 'weighted', 'envelope', 'gaussian']:
            mask = gs.data.l == 0
            mask = np.logical_and(mask, gs.data.n_pg > 0)

            delta_nus = np.diff(nu_all[mask])
            if delta_nu == 'mean':
                delta_nu = np.mean(delta_nus)
            elif delta_nu == 'median':
                delta_nu = np.median(delta_nus)
            elif delta_nu in ['weighted', 'envelope', 'gaussian']:
                delta_nu = uf.calc_deltanu(gs, hist)

    mask = gs.data.l == 0
    E_l0 = gs.get('E_norm')[mask]
    nu_l0 = nu_all[mask]
    # interpolate over log10 inertia for better behaviour
    try:
        log_f_El0 = interp1d(nu_l0, np.log10(E_l0), kind='cubic', bounds_error=True)
    except ValueError:
        log_f_El0 = interp1d(nu_l0, np.log10(E_l0), kind='linear', bounds_error=True)
    if offset == 'auto_old':
        # Find an offset which results in the minimum amount of wrapping in the echelle diagram.
        def get_length(x):
            return np.sum(np.abs(np.diff(x)))

        lengths = []
        for offset in np.linspace(0, delta_nu, int(delta_nu * 10) + 1):
            length = 0

            for l in l_list:
                mask = gs.data.l == l

                nu = nu_all[mask]
                mod_nu = (nu + offset) % delta_nu

                length += get_length(mod_nu)

            lengths.append([offset, length])
        lengths = np.array(lengths)
        offset = np.where((lengths[:, 1] - lengths[:, 1].min()) < delta_nu * 1e-3)[0][0]

    elif offset == 'auto':
        nu_mod_dnu = nu_all[gs.data.l == 0] % delta_nu
        if (np.max(nu_mod_dnu) - np.min(nu_mod_dnu)) > 0.5 * delta_nu:  # has a fold
            nu_mod_dnu[nu_mod_dnu <= 0.5 * delta_nu] += delta_nu

        offset = 0.2 * delta_nu - np.min(nu_mod_dnu)
        offset = offset % delta_nu

    for i, l in enumerate(l_list):
        mask = gs.data.l == l

        nu = nu_all[mask]
        ms = pu.calc_inertia_marker_size(gs, l=l, freq_units=freq_units)
        ax.scatter((nu + offset) % delta_nu, nu, ms, c=f'C{i}', label=r'$\ell={}$'.format(l),
                   marker=['o', 's', '^', 'v'][i % 4])

    freq_unit = {'uHz': r'\mu', 'mHz': 'm', 'Hz': ''}[freq_units]
    ax.set_xlabel(r'Frequency + {:.2f} mod {:.2f} $\mu$Hz'.format(offset, delta_nu))
    ax.set_ylabel(f'Frequency (${freq_unit}$Hz)')
    if legend_loc != '':
        ax.legend(loc=legend_loc, ncol=len(l_list))

    ax.set_xlim(0, delta_nu)

    if hist is not None:
        nu_max = hist.get('nu_max')[hist_i]
        fmid = nu_max
        fsig = (0.66 * nu_max ** 0.88) / 2 / np.sqrt(2 * np.log(2.))  # Mosser 2012a

        fmin = max(1e-4, fmid - 2 * fsig - 0.5 * delta_nu)
        fmax = fmid + 2 * fsig + 0.5 * delta_nu

        ax.set_ylim(fmin, fmax)

    # ax.text(0, 0, rf'$\Delta\nu = {delta_nu:.2f}')

    return f, ax


#
# def make_period_echelle(gs, hist, ax=None,
#                  prefix='profile', suffix='.data.GYRE.sgyre_l', freq_units='uHz', legend_loc='upper right'):
#     f, ax = pu.get_figure(ax)
#
#     pnum = int(gs.path.split(prefix)[-1].replace(suffix, ''))
#     hist_i = hist.get_profile_index(pnum)[0]
#     nu_all = gs.get_frequencies(freq_units)
#     dPi1 = hist.get('delta_Pg')[hist_i]
#
#     ms = pu.calc_inertia_marker_size(gs, l=1, freq_units=freq_units)
#
#     nu = uf.get_freq(gs, 'Hz')
#
#     # mask = None
#     # ax.scatter(1/nu[mask] % dPi1, nu, ms, color='C1', label=fr'$\ell=1$')
#
#     freq_unit = {'uHz': r'\mu', 'mHz': 'm', 'Hz': ''}[freq_units]
#     ax.set_xlabel(r'$1/\nu \; \mathrm{mod} \; \Delta \Pi_1$')
#     ax.set_ylabel(f'Frequency (${freq_unit}$Hz)')
#
#     if legend_loc != '':
#         ax.legend(loc=legend_loc, ncol=len(l_list))
#     ax.scatter((1 / nu % dPi1) - dPi1, nu, ms, color='C1')
#     ax.axvline(0, 0, 1, ls='--', color='k', zorder=-3)
#     return f, ax


def make_inertia(gs, ax=None, l_list=(0, 1, 2), freq_units='uHz', div=True, legend_loc='upper right',
                 scale_marker=False):
    f, ax = pu.get_figure(ax)
    freqs = gs.get_frequencies(freq_units)

    mask = gs.get('l') == 0
    if div:
        f_min_l0 = freqs[mask].min()
        f_max_l0 = freqs[mask].max()
    else:
        f_min_l0 = 0
        f_max_l0 = 1e99
    E_l0 = gs.get('E_norm')[mask]
    # interpolate over log10 inertia for better behaviour
    try:
        log_f_El0 = interp1d(freqs[mask], np.log10(E_l0), kind='cubic', bounds_error=True)
    except ValueError:
        log_f_El0 = interp1d(freqs[mask], np.log10(E_l0), kind='linear', bounds_error=True)

    for i, l in enumerate(l_list):
        mask = (gs.get('l') == l) & (freqs >= f_min_l0) & (freqs <= f_max_l0)
        E = gs.get('E_norm')[mask]
        if scale_marker:
            ms = pu.calc_inertia_marker_size(gs, l, freq_units)
        if div:
            E_div_El0 = E / 10 ** log_f_El0(freqs[mask])
            if scale_marker:
                ax.scatter(freqs[mask], E_div_El0, s=ms, color=f'C{l}', label=fr'$\ell={l}$',
                           marker=['o', 's', '^', 'v'][i % 4])
                ax.plot(freqs[mask], E_div_El0, color=f'C{l}', lw=1)
            else:
                ax.plot(freqs[mask], E_div_El0, label=fr'$\ell={l}$', lw=1, color=f'C{l}',
                        marker=['o', 's', '^', 'v'][i % 4])
        else:
            if l == 0:
                x = np.linspace(min(freqs[mask]), max(freqs[mask]), 1000)
                ax.plot(x, 10 ** log_f_El0(x), 'C0-', lw=1)
                if scale_marker:
                    ax.scatter(freqs[mask], E, s=ms, color=f'C{l}', label=fr'$\ell={l}$',
                               marker=['o', 's', '^', 'v'][i % 4])
                else:
                    ax.plot(freqs[mask], E, 'C0o', lw=1)
                    ax.plot([], [], 'C0', lw=1, label=fr'$\ell={l}$', marker=['o', 's', '^', 'v'][i % 4])
            else:
                if scale_marker:
                    ax.scatter(freqs[mask], E, s=ms, color=f'C{l}', label=fr'$\ell={l}$',
                               marker=['o', 's', '^', 'v'][i % 4])
                    ax.plot(freqs[mask], E, color=f'C{l}', lw=1)
                else:
                    ax.plot(freqs[mask], E, label=fr'$\ell={l}$', lw=1, color=f'C{l}',
                            marker=['o', 's', '^', 'v'][i % 4])

    if legend_loc != '':
        ax.legend(loc=legend_loc, ncol=len(l_list))

    freq_unit = {'uHz': r'\mu', 'mHz': 'm', 'Hz': ''}[freq_units]
    ax.set_xlabel(f'Mode frequency  (${freq_unit}$Hz)')
    if div:
        ax.set_ylabel(f'Normalized mode inertia $E_{{\\ell}} / E_{{\\ell=0}}$')
    else:
        ax.set_ylabel(f'Mode inertia $E_{{\\ell}}$')
    ax.set_yscale('log')

    return f, ax


def make_resolutions(profs, ax=None, yres=100, xname='model_number', yname='mass', zorder=0):
    f, ax = pu.get_figure(ax)

    n_zones = 0
    for p in profs:
        n_zones += p.header['num_zones']
    dat = np.zeros((n_zones, 2))
    i = 0

    x_edges = [0]
    for p in profs:
        nz = p.header['num_zones']
        dat[i:i + nz, 0] = p.header[xname]
        dat[i:i + nz, 1] = p.get(yname)
        i += nz

        x_edges.append(p.header[xname])
    x_edges = np.array(x_edges)
    x_edges[0] = x_edges[1] - (x_edges[2] - x_edges[1])

    counts, edges_x, edges_y, cax = ax.hist2d(dat[:, 0], dat[:, 1],
                                              bins=((x_edges),
                                                    np.linspace(dat[:, 1].min(), dat[:, 1].max(), yres + 1)),
                                              norm=colors.LogNorm(vmin=1), cmin=1, zorder=zorder)
    f.colorbar(cax, ax=ax)

    return f, ax, (counts, edges_x, edges_y)


def make_hrd_profiles(hist, add_cbar=True, vHRD_norm=1.0, use_mask=True, ax=None, show_prof_num=-1):
    f, ax = make_vhrd(hist, add_cbar, vHRD_norm, use_mask, ax)

    mask = uf.get_mask(hist, use_mask)
    masked_modnum = hist.get('model_number')[mask]

    model_num, _, pnum = hist.index.T
    model_num = np.intersect1d(model_num, masked_modnum)
    hist_i = np.where(np.isin(hist.data.model_number, model_num))[0]

    xdat, ydat = uf.get_logTeffL(hist)

    ax.plot(xdat[hist_i], ydat[hist_i], 'k+')

    if show_prof_num > 0:
        for i in range(0, len(hist_i), show_prof_num):
            ax.text(xdat[hist_i[i]], ydat[hist_i[i]], str(pnum[i]), verticalalignment='top')

    return f, ax


def make_hrd_models(hist, add_cbar=True, vHRD_norm=1.0, use_mask=True, ax=None, show_mod_num=1000, bounds=None):
    f, ax = make_vhrd(hist, add_cbar, vHRD_norm, use_mask, ax, bounds)

    mask = uf.get_mask(hist, use_mask)
    masked_modnum = hist.get('model_number')[mask]

    hist_i = np.where(np.logical_and(hist.data.model_number % show_mod_num == 0,
                                     np.isin(hist.data.model_number, masked_modnum)))[0]

    xdat, ydat = uf.get_logTeffL(hist)
    ax.plot(xdat[hist_i], hist.data.log_L[hist_i], 'k+')

    if show_mod_num > 0:
        for i in range(0, len(hist_i)):
            ax.text(xdat[hist_i[i]], hist.data.log_L[hist_i[i]], str((i + 1) * show_mod_num),
                    verticalalignment='top')

    return f, ax


def make_age_nu(hist, gss, l=1, ax=None, gyre_summary_prefix='profile',
                gyre_summary_suffix='.data.GYRE.sgyre_l', age_unit='Myr', marker='k.'):
    f, ax = pu.get_figure(ax)

    pnums = []
    for gs in gss:
        fname = os.path.split(gs.path)[-1]
        if fname.startswith(gyre_summary_prefix) and fname.endswith(gyre_summary_suffix):
            pnum = int(fname[len(gyre_summary_prefix):-len(gyre_summary_suffix)])
        else:
            raise ValueError('Gyre summary file {} does not '
                             'start with `{}` or end with `{}`.'.format(gs.path, gyre_summary_prefix,
                                                                        gyre_summary_suffix))
        pnums.append(pnum)

    asort = np.argsort(pnums)
    gss = [gss[i] for i in asort]
    pnums = [pnums[i] for i in asort]

    age_factor = {'yr': 1e0, 'kyr': 1e3, 'Myr': 1e6, 'Gyr': 1e9}[age_unit]

    hist_i = hist.get_profile_index(pnums)
    ages = hist.get('star_age')[hist_i] / age_factor

    for i, gs in enumerate(gss):
        mask = gs.get('l') == l
        mask = np.logical_and(mask, gs.data.n_pg > 0)
        nu = gs.get_frequencies('uHz')[mask]

        age = ages[i] * np.ones_like(nu)
        ax.plot(age, nu, marker)

    ax.set_xlabel('Age ${}$'.format(age_unit))
    ax.set_ylabel(r'Frequency ($\mu$Hz)')

    return f, ax


def make_mesa_gyre_delta_nu(hist, gss, l_list=(0, 1, 2), xaxis='model_number', gyre_summary_prefix='profile',
                            gyre_summary_suffix='.data.GYRE.sgyre_l', ax=None):
    if ax is None:
        f, (ax0, ax1) = plt.subplots(nrows=2, ncols=1, sharex=True, gridspec_kw={'height_ratios': [3, 1]})

    elif not hasattr(ax, '__len__'):
        f = pu.get_figure()
        bbox = ax.get_position()
        width, height = bbox.p1 - bbox.p0

        bbox2_p0 = bbox.p0.copy()
        bbox2_p1 = bbox.p1.copy()
        bbox2_p0[1] += height / 3
        bbox2 = transforms.Bbox([bbox2_p0, bbox2_p1])

        ax.set_position(bbox2)
        ax.xaxis.set_visible(False)
        ax0 = ax

        bbox3_p0 = bbox.p0.copy()
        bbox3_p1 = bbox.p1.copy()
        bbox3_p1[1] = bbox2_p0[1]
        bbox3 = transforms.Bbox([bbox3_p0, bbox3_p1])
        ax1 = f.add_axes(bbox3)

    else:
        ax0, ax1 = ax
        f = pu.get_figure()

    ax0.tick_params(labelbottom=False)

    gyre_delta_nus = []

    for gs in gss:
        pnum = int(gs.fname[len(gyre_summary_prefix):-len(gyre_summary_suffix)])
        delta_nus = []
        for l in l_list:
            mask = gs.get('l') == l
            mask = np.logical_and(mask, gs.data.n_pg > 0)

            delta_nus.extend(np.diff(gs.get_frequencies('uHz')[mask]))

        median_delta_nu = np.mean(delta_nus)
        mean_delta_nu = np.median(delta_nus)

        gyre_delta_nus.append([pnum, median_delta_nu, mean_delta_nu])
    gyre_delta_nus = np.array(gyre_delta_nus)
    hist_i = hist.get_profile_index(gyre_delta_nus[:, 0])

    xdat = hist.get(xaxis)
    mesa_delta_nu = hist.get('delta_nu')

    ax0.plot(xdat[hist_i], gyre_delta_nus[:, 1], 'xC1', label='median')
    ax0.plot(xdat[hist_i], gyre_delta_nus[:, 2], '+C2', label='mean')
    ax0.plot(xdat, mesa_delta_nu, label='MESA')
    ax0.set_ylabel(r'$\Delta \nu$ ($\mu$Hz)')
    # ax0.set_ylim([0, ax0.get_ylim()[1]])
    ax0.set_yscale('log')

    ax0.legend()

    ax1.plot(xdat[hist_i], gyre_delta_nus[:, 1] / mesa_delta_nu[hist_i], 'xC1', label='median')
    ax1.plot(xdat[hist_i], gyre_delta_nus[:, 2] / mesa_delta_nu[hist_i], '+C2', label='mean')
    ax1.grid(axis='y')

    ax1.set_ylabel(r'$\Delta \nu_{\mathrm{gyre}} / \Delta\nu_{\mathrm{MESA}}$')
    ax1.set_xlabel(xaxis)

    # ax0.set_xlim([min(xdat), max(xdat)])
    # ax1.set_xlim([min(xdat), max(xdat)])

    return f, (ax0, ax1)


def make_composition(prof, elements='all', xname='mass', ax=None, add_burnmix=True, legend_loc='best', add_cbar=True,
                     add_legend=False, add_label_lines=True, num_line_label=4, normalize_comp=False, grad=False,
                     fontsize=7, hist=None):
    f, ax = pu.get_figure(ax)

    e_pp = ['h1', 'he4']
    e_cno = ['h1', 'he4', 'c12', 'n14', 'o16']
    e_3a = ['he4', 'c12', 'o16']
    e_c = ['he4', 'c12', 'o16', 'ne20', 'mg24']

    do_xyz = False
    if isinstance(elements, str):
        elements = elements.lower()
    if elements == 'all':
        elems = [_ for _ in prof.columns if len(_) <= 4 and not _.startswith('log') and _[-1].isdigit()]
    elif elements == 'pp':
        elems = e_pp
    elif elements == 'cno':
        elems = e_cno
    elif elements == '3a':
        elems = e_3a
    elif elements == 'c':
        elems = e_c
    elif elements == 'xyz':
        do_xyz = True
        elems = ['X', 'Y', 'Z']
        X = prof.get('h1')
        if 'h2' in prof.columns:
            X += prof.get('h2')
        if 'h3' in prof.columns:
            X += prof.get('h3')
        Y = prof.get('he4')
        if 'he3' in prof.columns:
            Y += prof.get('he3')
        Z = 1 - X - Y
    else:
        elems = elements

    xdata = pu.get_x_and_set_xlabel(prof, xname, ax=ax, hist=hist)

    _colors = ['C0', 'C1', 'C2', 'C3', 'C5', 'C7', 'C8', 'C9']
    linestyles = ['-', '--', '-.', ':']
    for i, e in enumerate(elems):
        _c = _colors[i % len(_colors)]
        ls = linestyles[i // len(_colors)]
        if do_xyz:
            if e == 'X':
                comp = X
            elif e == 'Y':
                comp = Y
            elif e == 'Z':
                comp = Z
        else:
            comp = prof.get(e)
        if normalize_comp:
            comp -= min(comp)
            comp /= max(comp)
        if grad:
            if 'pressure' in prof.columns:
                pressure = prof.get('pressure')
            elif 'logP' in prof.columns:
                pressure = 10**prof.get('logP')
            else:
                raise KeyError('No column name with name pressure or logP.')
            comp = uf.cell2face(comp, prof.data.mass, True)
            comp_grad = np.gradient(comp, pressure) / (comp / pressure)
            comp_grad[comp < 1e-9] = np.nan
            comp = np.abs(comp_grad)

        ax.plot(xdata, comp, c=_c, ls=ls, label=e.capitalize())

    if grad:
        ax.set_ylabel('abs dlnX/dlnP')
    else:
        ax.set_ylabel('Mass fraction')
        ax.set_yscale('log')
    if xname in ['logR']:
        ax.set_xscale('log')

    if xname == 'mass':
        ax.set_xlim(0, None)
    if not grad:
        if do_xyz:
            ax.set_ylim(10 ** -3.5, 2)
            ax.set_yticks([1e0, 1e-1, 1e-2, 1e-3])
        else:
            ax.set_ylim(10 ** -6.5, 2)
            ax.set_yticks([1e0, 1e-3, 1e-6])

    if add_burnmix == 'mix':
        pu.add_mixing(ax, prof, xname, add_legend=False)
    elif add_burnmix == 'burn':
        pu.add_burning(ax, prof, xname, add_cbar=add_cbar)
    else:
        if add_burnmix:
            pu.add_burning(ax, prof, xname, add_cbar=add_cbar)
            pu.add_mixing(ax, prof, xname, add_legend=False)

    if add_label_lines:
        pu.line_legend(ax, num_line_label=num_line_label, fontsize=fontsize)

    if add_legend:
        ax.legend(loc=legend_loc)
    return f, ax


def make_gradients(prof, xname='mass', hist=None, ax=None, add_legend=True, n_cols_legend=2):
    f, ax = pu.get_figure(ax)

    x = pu.get_x_and_set_xlabel(prof, xname, ax=ax, hist=hist)
    labels = {'gradr': r'$\nabla_\mathrm{rad}$', 'grada': r'$\nabla_\mathrm{ad}$',
              'actual_gradT': r'actual $\nabla$', 'gradT': r'$\nabla$'}

    for i, grad in enumerate(['gradr', 'grada', 'actual_gradT', 'gradT']):
        try:
            ax.plot(x, prof.get(grad), ['k-', 'k--', 'r', 'r'][i], label=labels[grad], zorder=-(i // 2) + 2,
                    lw=1.5 + 4 * (i // 2))
        except ValueError:
            pass

    if 'logP' in prof.columns:
        logP = prof.get('logP')
        skip_grad_mu = False
    elif 'pressure' in prof.columns:
        logP = np.log10(prof.get('pressure'))
        skip_grad_mu = False
    else:
        skip_grad_mu = True

    if 'mu' in prof.columns:
        pass
    else:
        skip_grad_mu = False

    if not skip_grad_mu:
        grad_mu = np.gradient(np.log10(prof.get('mu'), logP))
        ax.plot(x, grad_mu, 'k-.', label=r'$\nabla_\mu$')

    if add_legend:
        pu.top_legend(ax, n_cols_legend)
    ax.set_ylim(0, 1)
    return f, ax


def make_period_spacing(gs, hist=None, ax=None, freq_units='uHz', prefix='profile', suffix='.data.GYRE.sgyre_l',
                        l_list=(0, 1, 2), legend_loc='upper right'):
    f, ax = pu.get_figure(ax)

    if hist is not None:
        pnum = int(gs.path.split(prefix)[-1].replace(suffix, ''))
        hist_i = hist.get_profile_index(pnum)[0]
        ax.axvline(hist.get('nu_max')[hist_i], color='k', zorder=-3)
    for i, l in enumerate(l_list):
        if l == 0:  # No radial g-modes
            continue
        mask = gs.get('l') == l
        nu = gs.get_frequencies(freq_units)[mask]
        dPi = -np.diff((nu * {'uHz': 1e-6, 'mHz': 1e-3, 'Hz': 1}[freq_units]) ** -1)

        # ax.plot(dPi, nu[:-1] % hist.get('delta_Pg')[hist_i], f'C{l}.', label=fr'$\ell={l}$')

        ms = pu.calc_inertia_marker_size(gs, l=l, freq_units=freq_units)

        # ax.plot(nu[:-1], dPi, f'C{l}.', label=fr'$\ell={l}$')
        ax.scatter(nu[:-1], dPi, s=ms[:-1], color=f'C{l}', marker=['o', 's', '^', 'v'][i % 4])
        ax.plot(nu[:-1], dPi, color=f'C{l}', lw=1)
        line = plt.matplotlib.lines.Line2D([], [], linewidth=1, linestyle='-', marker=['o', 's', '^', 'v'][i % 4],
                                           color=f'C{l}', label=fr'$\ell={l}$')
        ax.add_line(line)

    freq_prefix = {'uHz': r'\mu', 'mHz': 'm', 'Hz': ''}[freq_units]
    ax.set_ylabel(r'$\Delta P$ (s)')
    ax.set_xlabel(f'Mode frequency  (${freq_prefix}$Hz)')

    if legend_loc != '':
        ax.legend(loc=legend_loc)
    # ax.set_xlim(0, None)

    return f, ax


def make_structural(prof, ax=None, xname='mass', hist=None, normalize=False):
    f, ax = pu.get_figure(ax)

    x = pu.get_x_and_set_xlabel(prof, xname, ax, hist=hist)

    for yname in ('logP', 'logRho', 'logT', 'logR'):
        if yname in prof.columns:
            ydata = prof.get(yname)
        else:
            other_yname = {'logP': 'pressure', 'logRho': 'density', 'logT': 'temperature', 'logR': 'radius'}[yname]
            ydata = np.log10(prof.get(other_yname))
        if normalize:
            ydata -= min(ydata)
            ydata /= max(ydata)
        ax.plot(x, ydata, label=yname)

    if xname == 'mass':
        ax.set_xlim(0, None)

    pu.line_legend(ax, edge_space=0.075)
    ax.set_xlabel(xname)

    return f, ax


# def make_eos(prof, ax=None, mesa_version='15140', cbar_min=0, cbar_max=None):
#     f, ax = pu.get_figure(ax)
#
#     eos_path = os.path.join(os.path.dirname(__file__), f'../misc/eos_plotter_{mesa_version}.dat')
#
#     # Based on plotter.py from $MESA_DIR/eos
#
#     def parse(fname):
#         nY, nX = np.loadtxt(fname, max_rows=1, skiprows=3, unpack=True, dtype=int)
#         data = np.loadtxt(fname, skiprows=4)
#         data = np.reshape(data, (nX, nY, -1))
#         Yran = data[0, :, 0]
#         Xran = data[:, 0, 1]
#         data = np.swapaxes(data, 0, 1)
#         return data, Yran, Xran
#
#     with open(eos_path, 'r') as handle:
#         title = handle.readline().strip()
#         xlabel = handle.readline().strip()
#         ylabel = handle.readline().strip()
#
#     eosDT, Yran, Xran = parse(eos_path)
#
#     # set up plot and labels
#     ax.set_title(title)
#     ax.set_xlabel(xlabel)
#     ax.set_ylabel(ylabel)
#     ax.set_xlim(Xran.min(), Xran.max())
#     ax.set_ylim(Yran.min(), Yran.max())
#
#     # set up color map
#     cmap = copy.copy(mpl.cm.get_cmap("viridis"))
#     cmap.set_over('white')
#     cmap.set_under('black')
#
#     pcol = ax.pcolormesh(Xran, Yran, eosDT[..., 2], shading='nearest', cmap=cmap, vmin=cbar_min, vmax=cbar_max)
#     pcol.set_edgecolor('face')
#     cax = f.colorbar(pcol, extend='both')
#     cax.set_label('')
#
#     try:
#         xprof = prof.get('logRho')
#     except ValueError:
#         xprof = np.log10(prof.get('density'))
#
#     try:
#         yprof = prof.get('logT')
#     except ValueError:
#         yprof = np.log10(prof.get('temperature'))
#     ax.plot(xprof, yprof, 'k')
#
#     return f, ax


def make_eigenfunc_compare(gs, gefs, prof, hist, l_list=(1,), prop_y_lims=(3e0, 5e4), eig_y_lims=None,
                           x_lims=(1e-3, 0.5), prop_kwargs={}, inertia_kwargs={}):
    f, axes = plt.subplot_mosaic([list('AAADD'), list('BBBDD'), list('CCC..')])
    axes['A'].sharex(axes['B'])
    axes['B'].sharex(axes['C'])
    i_hist = prof.get_hist_index(hist)
    if eig_y_lims is None:
        eig_y_lims = ((None, None), (None, None))

    if len(eig_y_lims) == 1:
        axes['B'].get_shared_y_axes().join(axes['B'], axes['C'])
    axes['D'].yaxis.set_label_position('right')
    axes['D'].yaxis.tick_right()

    prop_kwargs_use = {'xname': 'x', 'ax': axes['A'], 'add_legend': False}
    prop_kwargs_use.update(prop_kwargs)
    make_propagation(prof, hist, **prop_kwargs_use)

    inertia_kwargs_use = {'ax': axes['D'], 'l_list': l_list, 'div': False, 'legend_loc': ''}
    inertia_kwargs_use.update(inertia_kwargs)
    make_inertia(gs, **inertia_kwargs_use)
    axes['D'].axvline(hist.get('nu_max')[i_hist], color='k', zorder=-5)

    if 'density' in prof.columns:
        rho = prof.get('density')
    elif 'logRho' in prof.columns:
        rho = 10 ** prof.get('logRho')
    scaling_func = interp1d(prof.data.radius / prof.data.radius[0], prof.data.radius * rho ** 0.5, bounds_error=False)

    for i, gef in enumerate(gefs):
        scaling = scaling_func(gef.data.x)
        axes['A'].axhline(gef.header['Re(freq)'], color=f'C{i % 10}')
        axes['B'].plot(gef.data.x, scaling * gef.get('Re(xi_r)'), color=f'C{i % 10}')
        axes['C'].plot(gef.data.x, scaling * gef.get('Re(xi_h)'), color=f'C{i % 10}')

        l = gef.header['l']
        n_pg = gef.header['n_pg']
        n_p = gef.header['n_p']
        mask = (gs.get('l') == l) & (gs.data.n_pg == n_pg) & (gs.data.n_p == n_p)
        intertia = gs.data.E_norm[mask]
        axes['D'].plot(gs.get('Re(freq)')[mask], intertia, f'kX', zorder=9)
        axes['D'].plot(gs.get('Re(freq)')[mask], intertia, f'C{i % 10}x', zorder=10)

    axes['A'].set_ylim(prop_y_lims)
    if len(eig_y_lims) == 1:
        axes['B'].set_ylim(eig_y_lims[0])
    else:
        axes['B'].set_ylim(eig_y_lims[0])
        axes['C'].set_ylim(eig_y_lims[1])
    axes['B'].set_xlim(x_lims)

    axes['C'].set_xlabel(axes['A'].get_xlabel())
    axes['A'].set_xlabel('')

    axes['B'].set_ylabel(r'$r \; \rho^{1/2} \; \xi_r$')
    axes['C'].set_ylabel(r'$r \; \rho^{1/2} \; \xi_h$')

    axes['A'].set_xscale('log')
    # axes['A'].xaxis.set_ticklabels([])
    axes['A'].xaxis.set_ticklabels([], minor=False)
    axes['B'].xaxis.set_ticklabels([], minor=False)
    axes['A'].xaxis.set_ticklabels([], minor=True)
    axes['B'].xaxis.set_ticklabels([], minor=True)

    return f, axes

def make_kipp(hist, profs=None, ax=None, xaxis='model_number', yaxis='mass', caxis='eps_net', zone_filename='zones_wsssss.dat',
                 verbose=False, save_zones=True, clobber_zones=False, prof_prefix='profile', prof_suffix='.data', mixing_min_height=0,
                 prof_resolution=500, xlims=None, ylims=None, clims=None, kwargs_mixing=None, norm=None, cmap=None,
                 kwargs_profile_color=None, return_Kipp_data=False, parallel=True, ignore_monotonic=False, legend_loc=None, add_cbar=False):
    """
    Create a Kippenhahn diagram.

    Args:
        hist (History): MESA History with which to create a Kippenhahn diagram.
        profs (list of Profile, optional): MESA Profile used to create a Kippenhahn diagram. Is optional if `caxis='eps_net'` and `burning_regions` in the history columns is used.
        ax (matplotlib axes, optional): Matplotlib figure axis. If set to `None` will create one.
        xaxis (str, optional): x-axis column name. Defaults to 'model_number'.
        yaxis (str, optional): y-axis column name. Must be one of `mass` or `radius`. Defaults to `mass`.
        caxis (str, optional): colour-axis column name. Must be 'eps_net' or a column in `Profile`.
        zone_filename (str, optional): Filename of zonefile.
        verbose (bool, optional): If ``True``, prints extra information.
        save_zones (bool, optional): If ``True``, will cache Kippenhahn regions in ``zone_filename``.
        clobber_zones (bool, optional): If ``True``, will generate Kippenhahn regions from scratch.
        prof_prefix (str, optional): Profile file name prefix.
        prof_suffix (str, optional): Profile file name suffix.
        prof_resolution (int, optional): y-axis resolution for Profiles.
        mixing_min_height (float, optional): Minimum fractional vertical extent of mixing region. If smaller, the region is not plotted.
        xlims (length 2 array, optional): Lower and upper limits of the x-axis.
        ylims (length 2 array, optional): Lower and upper limits of the y-axis.
        clims (length 2 array, optional): Lower and upper limits of the colour-axis. Cannot be used at the same time with norm.
        kwargs_mixing (dict, optional): kwargs used to draw mixing regions, if `None`, defaults to `plotting.utils.pu.get_default_mixing_kwargs()`.
        norm (optional): Matplotlib normalize class.
        cmap (optional): Matplotlib colormap.
        kwargs_profile_color (dict, optional): Passed to tripcolor in ``add_color``.
        return_Kipp_data (bool, optional): If ``True``, also returns Kipp_data.
        parallel (bool, optional): If ``True``, calculate Kippenhahn regions in parallel.
        legend_loc(str, optional): If set, create a legend along that side. Can be 'top' or 'right'.
        add_cbar(bool, optional): If ``True``, will create a colorbar for the color axis.

    Returns:
        tuple of matplotlib.figure.Figure, matplotlib.axes._axes.Axes, and optionally Kipp_data: Figure, axis, and optionally `Kipp_data` of the Kippenhahn diagram.

    """
    if clims is None and caxis == 'eps_net':
        clims = [-3, 8]

    if norm is not None:
        if kwargs_profile_color is not None:
            if 'norm' in kwargs_profile_color.keys():
                print(f'Using norm from argument.')
        else:
            kwargs_profile_color = {'norm': norm}
    else:
        if kwargs_profile_color is None:
            norm = None
        else:
            norm = kwargs_profile_color['norm']

    kd = kipper.Kipp_data(hist, profs, xaxis, yaxis, caxis, norm, zone_filename, verbose, save_zones, clobber_zones, prof_prefix,
                   prof_suffix, prof_resolution, parallel, ignore_monotonic)
    f, ax = kd.make_kipp(ax, xlims, ylims, clims, norm, cmap, mixing_min_height, kwargs_mixing=kwargs_mixing, kwargs_profile_color=kwargs_profile_color)

    ax.set_xlabel(xaxis.replace('_', ' '))
    ax.set_ylabel(yaxis)

    if legend_loc is not None:
        if kwargs_mixing is None:
            kwargs_mixing = pu.get_default_mixing_kwargs()
        labels = []
        handles = []
        for mixtype, val in kd.has_mixtype.items():
            if val:
                hatch = kwargs_mixing[mixtype]['hatch']
                color = kwargs_mixing[mixtype]['color']
                line = kwargs_mixing[mixtype]['line']
                show = kwargs_mixing[mixtype]['show']
                if show:
                    labels.append(uf.mix_dict['names'][uf.mix_dict['merged_r'][mixtype]])
                    handles.append(plt.Rectangle((0,0), 0, 0,
                                                 fill=False, hatch=hatch, edgecolor=color, linewidth=line))
        if len(labels) > 0:
            if legend_loc == 'top':
                legend = pu.top_figure_legend(f, 4, handles=handles, labels=labels)
            elif legend_loc == 'right':
                legend = pu.side_figure_legend(f, 1, handles=handles, labels=labels)
            else:
                pass

    if add_cbar and caxis:
        norm, cmap = kd.color_info

        f = ax.get_figure()
        f.colorbar(plt.cm.ScalarMappable(cmap=cmap, norm=norm), ax=ax)
        cbax = f.axes[-1]
        cbax.set_ylabel(caxis)

    if return_Kipp_data:
        return f, ax, kd
    return f, ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Numerical constants used by MESA in ``$MESA_DIR/const/public/const_def.f90``. This file is generated by ``wsssss/dev/create_const_files.py``.
"""

version = '11701'
"""MESA version used to generate this file."""

pi = 3.1415926535897932384626433832795028841971693993751E0
"""pi = 3.1415926535897932384626433832795028841971693993751E0"""

pi4 = 4*pi
"""pi4 = 4*pi"""

eulercon = 0.577215664901532861E0
"""eulercon = 0.577215664901532861E0"""

ln2 = 6.9314718055994529E-01
"""log_cr(2d0)"""

ln3 = 1.0986122886681096E+00
"""log_cr(3d0)"""

ln10 = 2.3025850929940455E+00
"""log_cr(10d0)"""

a2rad = pi/180.0E0
"""angle to radians"""

rad2a = 180.0E0/pi
"""radians to angle"""

one_third = 1E0/3E0
"""one_third = 1E0/3E0"""

two_thirds = 2E0/3E0
"""two_thirds = 2E0/3E0"""

ln4pi3 = 1.4324119583011810E0
"""log_cr(4*pi/3)"""

two_13 = 1.2599210498948730E0
"""pow_cr(2d0,1d0/3d0)"""

four_13 = 1.5874010519681994E0
"""pow_cr(4d0,1d0/3d0)"""

sqrt2 = 1.414213562373095E0
"""sqrt(2)"""

standard_cgrav  = 6.67428E-8
"""gravitational constant (g^-1 cm^3 s^-2)"""

planck_h  = 6.62606896E-27
"""Planck's constant (erg s)"""

hbar  = planck_h / (2*pi)
"""hbar  = planck_h / (2*pi)"""

qe  = 4.80320440E-10
"""electron charge (esu # (g cm^3 s^-2)^(1/2))"""

avo  = 6.02214179E23
"""Avogadro's constant (mole^-1)"""

clight  = 2.99792458E10
"""speed of light in vacuum (cm s^1)"""

kerg  = 1.3806504E-16
"""Boltzmann's constant (erg K^-1)"""

boltzm  = kerg
"""boltzm  = kerg"""

cgas  = boltzm*avo
"""R_gas# ideal gas constant# erg/K/mole"""

kev  = 8.617385E-5
"""converts temp (kelvin) to ev (ev K^-1)"""

amu  = 1.660538782E-24
"""atomic mass unit (g)"""

mn  = 1.6749286E-24
"""neutron mass (g)"""

mp  = 1.6726231E-24
"""proton mass (g)"""

me  = 9.1093826E-28
"""electron mass (g) (was 9.1093897d-28)"""

rbohr  = hbar*hbar/(me * qe * qe)
"""Bohr radius (cm)"""

fine  = qe*qe/(hbar*clight)
"""fine structure constant"""

hion  = 13.605698140E0
"""hydrogen ionization energy (eV)"""

ev2erg  = 1.602176487E-12
"""electron volt (erg)"""

mev_to_ergs  = 1E6*ev2erg
"""mev_to_ergs  = 1E6*ev2erg"""

mev_amu  = mev_to_ergs/amu
"""mev_amu  = mev_to_ergs/amu"""

qconv  = ev2erg*1.0E6*avo
"""convert Q rates to erg/gm/sec"""

boltz_sigma  = 5.670400E-5
"""boltzmann's sigma #  crad*clight/4 (erg cm^-2 K^-4 s^-1)"""

crad  = boltz_sigma*4/clight
"""radiation density constant, a (erg cm^-3 K^-4)# Prad  crad * T^4 / 3"""

sige = 6.6524587158E-025
"""Thomson scattering electron cross section"""

ssol  = boltz_sigma
"""ssol  = boltz_sigma"""

asol  = crad
"""asol  = crad"""

weinlam  = planck_h*clight/(kerg * 4.965114232E0)
"""weinlam  = planck_h*clight/(kerg * 4.965114232E0)"""

weinfre  = 2.821439372E0*kerg/planck_h
"""weinfre  = 2.821439372E0*kerg/planck_h"""

rhonuc  = 2.342E14
"""density of nucleus (g cm^3)"""

msol  = 1.9892E33
"""solar mass (g)"""

rsol  = 6.9598E10
"""solar radius (cm)"""

lsol  = 3.8418E33
"""solar luminosity (erg s^-1)"""

agesol  = 4.57E9
"""solar age (years)"""

msun  = msol
"""msun  = msol"""

rsun  = rsol
"""rsun  = rsol"""

lsun  = lsol
"""lsun  = lsol"""

msun33  = msol*1E-33
"""msun33  = msol*1E-33"""

rsun11  = rsol*1E-11
"""rsun11  = rsol*1E-11"""

lsun33  = lsol*1E-33
"""lsun33  = lsol*1E-33"""

teffsol  = 5777E0
"""temperature (k)"""

loggsol  = 4.4378893534131256E0
"""log surface gravity # log(g/(cm s^-2))"""

teffsun  = teffsol
"""teffsun  = teffsol"""

loggsun  = loggsol
"""loggsun  = loggsol"""

mbolsun  = 4.74
"""Bolometric magnitude of the Sun"""

mbolsol  = mbolsun
"""mbolsol  = mbolsun"""

ly  = 9.460528E17
"""light year (cm)"""

pc  = 3.261633E0 * ly
"""parsec (cm)"""

secyer = 3.1558149984E7
"""# seconds per year"""

dayyer = 365.25
"""# days per year"""

m_earth  = 5.9764E27
"""earth mass (g)"""

r_earth  = 6.37E8
"""earth radius (cm)"""

au  = 1.495978921E13
"""astronomical unit (cm)"""

m_jupiter  = 1.8986E30
"""jupiter mass (g)"""

r_jupiter  = 6.9911E9
"""jupiter mean radius (cm)"""

semimajor_axis_jupiter  = 7.7857E13
"""jupiter semimajor axis (cm)"""

arg_not_provided = -9E99
"""arg_not_provided = -9E99"""

missing_value = arg_not_provided
"""missing_value = arg_not_provided"""

no_mixing = 0
"""no_mixing = 0"""

convective_mixing = 1
"""convective_mixing = 1"""

softened_convective_mixing = 2
"""for modified D_mix near convective boundary"""

overshoot_mixing = 3
"""overshoot_mixing = 3"""

semiconvective_mixing = 4
"""semiconvective_mixing = 4"""

thermohaline_mixing = 5
"""thermohaline_mixing = 5"""

rotation_mixing = 6
"""rotation_mixing = 6"""

rayleigh_taylor_mixing = 7
"""rayleigh_taylor_mixing = 7"""

minimum_mixing = 8
"""minimum_mixing = 8"""

anonymous_mixing = 9
"""anonymous_mixing = 9"""




#!/usr/bin/env python

import unittest
import os
from .common import have_mesa_data

from wsssss import load_data as ld
from wsssss.plotting import plotting as pl
plt = pl.plt

have_mesa_data()
test_data = os.path.join(os.path.dirname(__file__), 'data', 'mesa')

class TestFunctions(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.hist = ld.History(f'{test_data}/0000/LOGS/history.data')
        cls.profs = ld.load_profs(cls.hist)
        cls.gss = ld.load_gss(cls.hist)
        

    def test_make_vhrd(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_vhrd(hist)
        plt.close('all')
        

    def test_make_hrd(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_hrd(hist)
        plt.close('all')


    def test_make_propagation(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_propagation(profs[-1], hist)
        plt.close('all')


    def test_make_propagation2(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_propagation2(profs[-1], hist)
        plt.close('all')


    def test_make_echelle(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_echelle(gss[-1], hist)
        f, ax = pl.make_echelle(gss[-1], None, delta_nu='median')
        plt.close('all')


    def test_make_inertia(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_inertia(gss[-1])
        plt.close('all')


    def test_make_resolutions(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax, (counts, edges_x, edges_y) = pl.make_resolutions(profs)
        plt.close('all')


    def test_make_hrd_profiles(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_hrd_profiles(hist)
        plt.close('all')


    def test_make_hrd_models(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_hrd_models(hist)
        plt.close('all')


    def test_make_age_nu(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_age_nu(hist, gss)
        plt.close('all')


    def test_make_mesa_gyre_delta_nu(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_mesa_gyre_delta_nu(hist, gss)
        plt.close('all')


    def test_make_composition(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_composition(profs[-1])
        plt.close('all')


    def test_make_gradients(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_gradients(profs[-1])
        plt.close('all')


    def test_make_period_spacing(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_period_spacing(gss[-1], hist)
        f, ax = pl.make_period_spacing(gss[-1], None)
        plt.close('all')


    def test_make_structural(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_structural(profs[-1])
        plt.close('all')


    # def test_make_eos(self):
    #     hist = self.hist
    #     profs = self.profs
    #     gss = self.gss
    #     f, ax = pl.make_eos(profs[-1])
    #     plt.close('all')


    def test_make_eigenfunc_compare(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        gef = ld.load_modes_from_profile(profs[-1])
        f, ax = pl.make_eigenfunc_compare(gss[-1], gef, profs[-1], hist)
        plt.close('all')


    def test_make_kipp(self):
        hist = self.hist
        profs = self.profs
        gss = self.gss
        f, ax = pl.make_kipp(hist)
        f, ax = pl.make_kipp(hist, profs, yaxis='mass')
        f, ax = pl.make_kipp(hist, profs, caxis='brunt_N2')
        norm = pl.pu.colors.LogNorm()
        f, ax = pl.make_kipp(hist, profs, caxis='brunt_N2', norm=norm)
        kwargs_colors = {'norm':norm}
        f, ax = pl.make_kipp(hist, profs, caxis='brunt_N2', kwargs_profile_color=kwargs_colors)
        f, ax = pl.make_kipp(hist, profs, caxis='brunt_N2', norm=norm, kwargs_profile_color=kwargs_colors)
        plt.close('all')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
This module contains various functions to load MESA and/or GYRE data.

Examples:
    Load a history file and the first profile:

    >>> import wsssss.load_data as ld
    >>> hist = ld.History('path/to/LOGS/history.data')
    >>> prof = ld.Profile('path/to/LOGS/profile1.data')
    >>> # or load all associated profiles:
    >>> profs = ld.load_profs(hist)

"""

import ast
import copy
import os
import shutil

import dill
import numpy as np

from . import functions as uf
from .constants import post15140
from .constants import pre15140


class _LazyProperty(object):
    def __init__(self, func):
        """
        Lazily load the attribute decorated by this class.
        Args:
            func: function
                Function which returns the property when called.
        """
        self._func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__

    def __get__(self, obj, klass=None):
        if obj is None:
            return None
        result = obj.__dict__[self.__name__] = self._func(obj)
        return result


class _Data:

    def __init__(self, path, keep_columns='all', save_dill=False, reload=False, verbose=False,
                 nanval=-1e99, nanclip=None, empty_on_error=False):
        """
        Common methods and attributes for History, Profile, and GyreSummary.

        Args:
            path (str): Path to the MESA history data file. If ending with `.dill`, will strip it and set that as `path`.
            keep_columns (list of str, optional): Which columns of the history data file to keep.
            save_dill (bool, optional): If True, will write a `.dill` file containing the `History` data.
            reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the history file.
            verbose (bool, optional): Print extra information.
            nanval (float, optional): Set all values equal to this to NaN.
            nanclip (2 floats, optional): Set all values outside this range to NaN.
        """
        self.path = os.path.abspath(path)
        self.directory = os.path.dirname(self.path)
        self.dill_only = False
        if os.path.isfile(self.path):
            self.fname = os.path.basename(self.path)
        elif os.path.isfile(self.path + '.dill') and not os.path.isfile(self.path):
            self.fname = os.path.basename(self.path)
            self.dill_only = True
        else:
            if empty_on_error:
                self.header = {}
                self.columns = []
                self.data = np.array([])
                self.loaded = False
                return
            else:
                raise FileNotFoundError(self.path)

        self.keep_columns = keep_columns
        self.save_dill = save_dill
        if self.path.endswith('.dill'):
            self.dill_path = path
        elif self.dill_only:
            self.dill_path = path + '.dill'
        else:
            self.dill_path = os.path.join(self.directory, self.fname + '.dill')

        self.loaded = False
        self.verbose = verbose
        self.nanval = nanval
        if nanclip is not None:
            self.nanclip = sorted(nanclip)
        else:
            self.nanclip = None

        if os.path.isfile(self.dill_path) and not reload:
            with open(self.dill_path, 'rb') as handle:
                try:
                    if not self.dill_only and (os.path.getmtime(self.dill_path) < os.path.getmtime(self.path)):
                        if self.verbose:
                            print('.dill file is older than loaded file! Reloading.')
                        self.save_dill = True
                        raise ValueError()
                    tmp = dill.load(handle)
                    self.header = tmp.header
                    self.columns = tmp.columns
                    self.data = tmp.data
                    self._first_row = tmp._first_row
                    if self.keep_columns != 'all':
                        self.columns = self.keep_columns
                        self.data = self._discard_columns_rec_array(self.data, self.columns)
                    self.loaded = True
                except Exception:
                    if self.verbose:
                        print("Failed to load dill from: \n{}".format(self.dill_path))
                    header, columns, first_row = self._read_data_file_header_columns()
                    self.header = header
                    self.columns = columns
                    self._first_row = first_row

        else:
            header, columns, first_row = self._read_data_file_header_columns()
            self.header = header
            self.columns = columns
            self._first_row = first_row

        if self.keep_columns != 'all':
            for col_keep in self.keep_columns:  # Check if columns present
                missing_cols = []
                if col_keep not in self.columns:
                    missing_cols.append(col_keep)
                if len(missing_cols) > 0:
                    raise ValueError(f'Columns in `keep_columns` not present in data file:\n'
                                     f'{" ".join(missing_cols)}')

    def __len__(self):
        return len(self.data)

    def __getitem__(self, mask):
        new_self = copy.copy(self)
        new_self.data = self._discard_rows_rec_array(self.data, mask)
        return new_self

    def _read_data_file_header_columns(self):
        """
        Read a data file and return the header, columns, and first row of data.

        Returns:
            A tuple (header, columns, first_line), where header is a dict containing the header of the data file,
            columns is a list of column names, and first_line a np.rec.array containing the first row of data.
        """
        lines = []
        with open(self.path, 'r') as handle:
            lines.extend(handle.readline() for _ in range(7))

        if lines[0].strip() == '':
            lines[:3] = lines[1:4]
            lines[3] = '\n'

        header_columns = lines[1].split()
        header_data = []
        for item in lines[2].split():
            try:
                val = ast.literal_eval(item)
            except ValueError:
                val = item
            header_data.append(val)
        header = {k: v for k, v in zip(header_columns, header_data)}
        if 'version_number' in header.keys():
            header['version_number'] = str(header['version_number'])

        columns = lines[5].split()
        first_row = lines[6].split()

        formats = [np.array(ast.literal_eval(_)).dtype if _ != 'NaN' else np.float64 for _ in first_row]
        first_line = np.rec.array(first_row, dtype=list(zip(columns, formats)))

        return header, columns, first_line

    def _fix_datafile(self):
        """
        Check for corrupted data in history. This can happen if, for example, MESA was killed when writing to the history.

        Returns:
            good_lines (list of str): Lines which do not have unexpected data.
        """
        with open(f'{self.path}', 'rb') as handle:
            lines = handle.readlines()

        expected_len = len(lines[6])

        bad_lines = []
        good_lines = []
        for i, line in enumerate(lines):
            if line.startswith(b'\x00'):
                bad_lines.append(i)
                bad_lines.append(i + 1)  # Line after line with \x00\x00... is garbled
            else:
                if len(line) != expected_len and i > 6:
                    bad_lines.append(i)
                else:
                    good_lines.append(line)
        bad_lines = [i for i in bad_lines if i >= 6]  # skip header for bad lines
        bad_lines = np.unique(bad_lines)

        print(f'Removed {len(bad_lines)} lines:\n{bad_lines}')
        return good_lines

    def _read_data_file(self):
        """
        Read a data file and return the header, columns, and data.

        Returns:
            A tuple (header, columns, data), where header is a dict containing the header of the data file,
            columns is a list of column names, and data a np.rec.array containing the data.
        """
        lines = []
        with open(self.path, 'r') as handle:
            lines.extend(handle.readline() for _ in range(7))

        if lines[0].strip() == '':
            lines[:2] = lines[1:3]
            lines[2] = '\n'

        header_columns = lines[1].split()
        header_data = [ast.literal_eval(_) for _ in lines[2].split()]
        header = {k: v for k, v in zip(header_columns, header_data)}

        columns = lines[5].split()
        first_row = lines[6].split()

        formats = [np.array(ast.literal_eval(_)).dtype if _ != 'NaN' else np.float64 for _ in first_row]
        try:
            data = np.rec.array(np.loadtxt(self.path, skiprows=6, dtype=list(zip(columns, formats))))
        except ValueError as exc:
            print(f"File {self.path} gave ValueError when reading:\n{exc.args[0]}\nTrying to fix.")

            shutil.copy2(self.path, f'{self.path}_original')

            lines = self._fix_datafile()

            with open(f'{self.path}', 'wb') as handle:  #
                handle.writelines(lines)

            data = np.rec.array(np.loadtxt(f'{self.path}', skiprows=6, dtype=list(zip(columns, formats))))

        return header, columns, data

    def _discard_columns_rec_array(self, rec_array, to_keep):
        """
        Recreate a record array from `rec_array` keeping only columns `to_keep`, and discarding other columns.

        Args:
            rec_array (np.rec.array): Record array from which to discard columns.
            to_keep (list of str): Columns names to keep.

        Returns:
            np.rec.array: New ``np.rec.array`` without discarded columns.
        """
        columns, formats = np.array(rec_array.dtype.descr).T
        mask = (columns != '') & np.isin(columns, to_keep)
        columns = columns[mask]
        formats = formats[mask]
        return np.rec.array(rec_array[columns].tolist(), dtype=list(zip(columns, formats)))

    def _discard_rows_rec_array(self, rec_array, mask):
        """
        Recreate a rec array from `rec_array` keeping only rows masked in `mask`, and discarding other rows.

        Args:
            rec_array (np.rec.array):
            mask (np.array): Mask to apply to column(s).

        Returns:
            np.rec.array: New record array without discarded rows.
        """
        columns, formats = np.array(rec_array.dtype.descr).T
        return np.rec.array(rec_array[mask].tolist(), dtype=list(zip(columns, formats)))

    @_LazyProperty
    def data(self):
        """
        Data loading function. Sets the ``header`` and ``column_names`` attributes and returns the main data.

        Returns:
            np.rec.array: Record array of the main data.
        """
        header, columns, data = self._read_data_file()

        if self.nanclip is not None:
            for col in columns:
                if data[col].dtype == np.float64:
                    data[col][data[col] <= self.nanclip[0]] = np.nan
                    data[col][data[col] >= self.nanclip[1]] = np.nan

        if self.nanval is not None:
            for col in columns:
                if data[col].dtype == np.float64:
                    data[col][data[col] == self.nanval] = np.nan

        self.data = data
        if isinstance(self, History):
            data = self._scrub_hist()

        self.data = data
        self.loaded = True

        if self.save_dill:
            self.dump()

        if self.keep_columns != 'all':
            self.columns = self.keep_columns
            self.data = self._discard_columns_rec_array(self.data, self.columns)
            data = self.data

        return data

    def dump(self, path_to_dump=''):
        """
        Dump a ``_Data`` object to disk as a ``.dill`` file.

        Args:
            path_to_dump (str, optional): Path where to write the file. If ``''``, will write at ``self.dill_path``.
        """
        if not self.loaded:  # Temporarily turn off save_dill so the lazy load of data won't write a dill file.
            save_dill = self.save_dill
            self.save_dill = False
            _ = self.data
            self.save_dill = save_dill

        if path_to_dump == '':
            path_to_dump = self.dill_path

        with open(path_to_dump, 'wb') as handle:
            dill.dump(self, handle)

    def get(self, *args, mask=None):
        """
        Get a single or multiple columns from ``data``.

        Args:
            *args (str): Column name(s) to get.
            mask (np.array): Mask to apply to column(s).

        Returns:
            ``np.rec.array`` or list of ``np.rec.array``: Column(s) of ``data``.

        """
        if mask is None:
            mask = ...

        if len(args) == 1:
            return self.data[args[0]][mask]
        else:
            return [self.data[cname][mask] for cname in args]


class _Mesa(_Data):

    def __init__(self, path, index_name='profiles.index', keep_columns='all', save_dill=False, reload=False,
                 verbose=False, nanval=-1e99, nanclip=None, empty_on_error=False):
        """
        Methods specific to History and Profile.

        Args:
            path (str): Path to the MESA history data file. If ending with `.dill`, will strip it and set that as `path`.
            index_name (str, optional): Filename of the profile index.
            keep_columns (list of str, optional): Which columns of the history data file to keep.
            save_dill (bool, optional): If True, will write a `.dill` file containing the `History` data.
            reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the history file.
            verbose (bool, optional): Print extra information.
            nanval (float, optional): Set all values equal to this to NaN.
            nanclip (2 floats, optional): Set all values outside this range to NaN.
        """
        super().__init__(path, keep_columns, save_dill, reload, verbose, nanval, nanclip, empty_on_error)

        self.LOGS = self.directory
        if os.path.isfile(index_name):
            self.index_path = os.path.abspath(index_name)
        else:
            self.index_path = os.path.join(self.LOGS, index_name)

        try:
            index = np.genfromtxt(self.index_path, skip_header=1, dtype=int)
            if index.shape == (3,):
                index = index.reshape((1, 3))

            # Scrub index of backups and retries
            max_model = index[-1, 0]
            index = index[index[:, 0] <= max_model]
            if isinstance(self, History):
                min_model = self._first_row.model_number
                index = index[index[:, 0] >= min_model]
            u, i = np.unique(index[:, 0][::-1], return_index=True)
            index = index[::-1][i]
        except OSError:
            if self.verbose:
                print('Index file not found, expected path {}'.format(self.index_path))
            index = None

        self.index = index


class History(_Mesa):
    def __init__(self, path, index_name='profiles.index', keep_columns='all', save_dill=True, reload=False,
                 verbose=False, nanval=-1e99, nanclip=None, empty_on_error=False):
        """
        Load a MESA history.

        Args:
            path (str): Path to the MESA history data file. If ending with `.dill`, will strip it and set that as `path`.
            index_name (str, optional): Filename of the profile index.
            keep_columns (list of str, optional): Which columns of the history data file to keep.
            save_dill (bool, optional): If True, will write a `.dill` file containing the `History` data. This file will
             include all columns, even those removed by keep_columns.
            reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the history file.
            verbose (bool, optional): Print extra information.
            nanval (float, optional): Set all values equal to this to NaN.
            nanclip (2 floats, optional): Set all values outside this range to NaN.
            empty_on_error (bool, optional): If loading fails, return an empty History.
        """
        super().__init__(path, index_name, keep_columns, save_dill, reload, verbose, nanval, nanclip, empty_on_error)

    def __getitem__(self, mask):
        new_self = copy.copy(self)
        new_self.data = self._discard_rows_rec_array(self.data, mask)
        if hasattr(self, 'index'):
            if self.index is not None:
                mnum0, mnum1 = new_self.data.model_number[[0, -1]]
                idx_mask = (self.index[:, 0] >= mnum0) & (self.index[:, 0] <= mnum1)
                new_self.index = self.index[idx_mask]
        return new_self

    def __repr__(self):
        initial_model = self._first_row['model_number']
        initial_mass = self._first_row['star_mass']
        initial_age = self._first_row['star_age']
        star_info = f'Initial model={initial_model}, mass={initial_mass:.2f}, age={initial_age}'
        return 'MESA history data file at {}'.format(self.path) + '\n' + star_info

    def get_profile_num(self, model_num, method='closest', earlier=True):
        """
        Returns the `closest` (by default) or `previous` or `next` profile number for a given model number.
        If earlier is True, and there are two closest profiles, return the one with a lower model number.

        Args:
            model_num (int):
            method (str): Must be one of `closest` or `previous`.
            earlier (bool):
        Returns:
            A tuple (header, columns, first_line), where header is a dict containing the header of the data file,
            columns is a list of column names, and first_line a np.rec.array containing the first row of data.
        Returns:
            A tuple (pnum, pmod, hist_ind), where pnum is the profile number, pmod is the model number of Profile, and
            hist_ind the index in History of Profile.
        """

        if method == 'closest':
            model_diff = np.abs(self.index[:, 0] - model_num)
            minval = min(model_diff)

            ind = np.where(model_diff == minval)[0]
            if len(ind) == 2:
                if earlier:
                    ind = ind[0]
                else:
                    ind = ind[1]
            else:
                ind = ind[0]
        elif method == 'previous':
            model_diff = self.index[:, 0] - model_num
            ind = np.where(model_diff <= 0)[0][-1]
        elif method == 'next':
            model_diff = self.index[:, 0] - model_num
            ind = np.where(model_diff >= 0)[0][0]
        else:
            raise ValueError("method must be 'closest' or 'previous' or 'next'.")
        pmod, _, pnum = self.index[ind]
        m_min, m_max = self.get('model_number')[[0, -1]]
        if (m_min <= pmod) and (m_max >= pmod):
            hist_ind = np.where(self.get('model_number') == pmod)[0][0]
        else:
            hist_ind = []
        return pnum, pmod, hist_ind


    def get_profile_index(self, profile_nums):
        """
        Returns the corresponding indeces of `profile_nums`.
        `profile_nums` can be an integer, a list of integers,
        a `Profile`, or a list of `Profile`s.

        Args:
            profile_nums (int or list of int): Profile numbers for which to calculate the indeces.

        Returns:
            int or np.array of int: Indeces of profile_nums in History.

        """
        if hasattr(profile_nums, '__len__'):
            if isinstance(profile_nums, Profile):
                profile_nums = [profile_nums]
        else:
            profile_nums = [profile_nums]
        if isinstance(profile_nums[0], Profile):
            profile_nums = [p.profile_num for p in profile_nums]
        idxs = np.where(np.isin(self.index[:, 2], profile_nums))
        model_nums = self.index[:, 0][idxs]
        data_idx = np.where(np.isin(self.data['model_number'], model_nums))[0]
        return data_idx

    def _scrub_hist(self):
        """
        Scrub history data for backups and retries.
        """
        max_model = self.data['model_number'][-1]
        scrubbed = self.data[self.data['model_number'] <= max_model]

        u, i = np.unique(scrubbed['model_number'][::-1], return_index=True)
        scrubbed = scrubbed[::-1][i]

        return scrubbed


class Profile(_Mesa):
    def __init__(self, path, index_name='profiles.index', keep_columns='all', load_GyreProfile=False,
                 suffix_GyreProfile='.GYRE', save_dill=False, reload=False, verbose=False, nanval=-1e99,
                 nanclip=None):
        """
        Load a MESA profile.

        Args:
            path (str): Path to the MESA profile data file. If ending with `.dill`, will strip it and set that as `path`.
            index_name (str, optional): Filename of the profile index.
            keep_columns (list of str, optional): Which columns of the history data file to keep.
            load_GyreProfile (bool): If True, will also load the corresponding GyreProfile into Profile.GyreProfile.
            suffix_GyreProfile (str, optional): Suffix of GyreProfile. Defaults to '.GYRE'.
            save_dill (bool, optional): If True, will write a `.dill` file containing the `History` data.
            reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the history file.
            verbose (bool, optional): Print extra information.
            nanval (float, optional): Set all values equal to this to NaN.
            nanclip (2 floats, optional): Set all values outside this range to NaN.
        """
        super().__init__(path, index_name, keep_columns, save_dill, reload, verbose, nanval, nanclip)
        self.LOGS = self.directory

        if self.index is not None:
            self.profile_num = self.index[np.where(self.index == self.header['model_number'])[0][0]][2]
        else:
            self.profile_num = None
        if load_GyreProfile:
            self.GyreProfile = GyreProfile(f'{self.path}{suffix_GyreProfile}')
        else:
            self.GyreProfile = None

    def __repr__(self):
        try_to_get = ['model_number', 'num_zones', 'star_mass', 'star_age', 'Teff', 'photosphere_L',
                      'center_h1', 'center_he4', 'date']
        return 'MESA profile data file at {}'.format(self.path) + '\n' + str(
            {key: self.header[key] for key in try_to_get if key in self.header.keys()})

    def get_hist_index(self, hist):
        """
        Get the index of this profile in the `History` hist.

        Args:
            hist (History):

        Returns:
            int: Index of profile in hist.
        """
        mod = self.header['model_number']
        return np.argwhere(hist.get('model_number') == mod)[0][0]


class _Gyre(_Data):

    def __init__(self, path, keep_columns='all', gyre_version='7', save_dill=False, reload=False, verbose=False,
                 nanval=-1e99, nanclip=None):
        """
        Common methods and attributes for GyreSummary and GyreMode.

        Args:
            path (str): Path to the Gyre summary or mode data file. If ending with `.dill`, will strip it and set that as `path`.
            keep_columns (list of str, optional): Which columns of the history data file to keep.
            gyre_version (str): Gyre version.
            save_dill (bool, optional): If True, will write a `.dill` file containing the `History` data.
            reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the history file.
            verbose (bool, optional): Print extra information.
            nanval (float, optional): Set all values equal to this to NaN.
            nanclip (2 floats, optional): Set all values outside this range to NaN.
        """
        super().__init__(path, keep_columns, save_dill, reload, verbose, nanval, nanclip)
        self.gyre_version = gyre_version


class GyreSummary(_Gyre):

    def __init__(self, path, keep_columns='all', gyre_version='7', save_dill=False, reload=False, verbose=False,
                 nanval=-1e99, nanclip=None):
        """
        Gyre summary output.

        Args:
            path (str): Path to the Gyre summary data file. If ending with `.dill`, will strip it and set that as `path`.
            keep_columns (list of str, optional): Which columns of the history data file to keep.
            gyre_version (str): Gyre version.
            save_dill (bool, optional): If True, will write a `.dill` file containing the `History` data.
            reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the history file.
            verbose (bool, optional): Print extra information.
            nanval (float, optional): Set all values equal to this to NaN.
            nanclip (2 floats, optional): Set all values outside this range to NaN.
        """
        super().__init__(path, keep_columns, gyre_version, save_dill, reload, verbose, nanval, nanclip)

    def __repr__(self):
        return f'GyreSummary at {self.path}'

    def _dimless_to_Hz(self):
        """
        Returns:
            float: Conversion factor between dimensionless frequency and Hz.

        """
        if 'M_star' in self.header.keys():
            M_star = self.header['M_star']
            R_star = self.header['R_star']
            G = pre15140.standard_cgrav
        else:
            M_star = self.get('M_star')[0]
            R_star = self.get('R_star')[0]
            G = post15140.standard_cgrav  # This changed in version 6.
        return 1.0 / (2 * np.pi) * ((G * M_star / (R_star) ** 3))

    def get_frequencies(self, freq_units, Re_freq_unit='uHz'):
        """
        Get frequencies in the specicied units. Will use 'Re(omega)' first and 'Re(freq)' otherwise.
        Args:
            freq_units (str): Unit to convert to, must be one of 'uHz', 'mHz', or 'Hz'.
            Re_freq_unit (str, optional): Unit of the 'Re(freq)' column, must be one of 'uHz', 'mHz', or 'Hz'.

        Returns:
            np.rec.array: Frequencies in unit specified by Re_freq_unit.
        """
        unit_dict = {'uHz': 1e6, 'mHz': 1e3, 'Hz': 1e0}
        if 'freq_units' in self.header.keys():
            header_freq_units = self.header['freq_units'].lower().replace('hz', 'Hz')
            if header_freq_units != Re_freq_unit:
                print(f'Warning! Frequency units in header ({header_freq_units}) and Re_freq_unit ({Re_freq_unit}) are '
                      f'not the same, using header frequency units.')
                Re_freq_unit = header_freq_units

        if 'Re(omega)' in self.columns:
            dimless_to_Hz = self._calc_dimless_to_Hz() * unit_dict[freq_units]
            freq_name = 'Re(omega)'
        elif 'Re(freq)' in self.columns:  # Assumes freq already in uHz.
            dimless_to_Hz = unit_dict[freq_units] / unit_dict[Re_freq_unit]
            freq_name = 'Re(freq)'
        return self.data[freq_name] * dimless_to_Hz


class GyreMode(_Gyre):

    def __init__(self, path, keep_columns='all', gyre_version='7', save_dill=False, reload=False, verbose=False,
                 nanval=-1e99, nanclip=None):
        super().__init__(path, keep_columns, gyre_version, save_dill, reload, verbose, nanval, nanclip)
        """
        Gyre mode detail file.

        Args:
            path (str): Path to the Gyre mode data file. If ending with `.dill`, will strip it and set that as `path`.
            keep_columns (list of str, optional): Which columns of the `GyreMode` data file to keep.
            gyre_version (str): Gyre version.
            save_dill (bool, optional): If True, will write a `.dill` file containing the `GyreMode` data.
            reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the mode file.
            verbose (bool, optional): Print extra information.
            nanval (float, optional): Set all values equal to this to NaN.
            nanclip (2 floats, optional): Set all values outside this range to NaN.
        """

    def __repr__(self):
        return f'GyreMode at {self.path}'

    def _dimless_to_Hz(self):
        """
        Returns:
            float: Conversion factor between dimensionless frequency and Hz.

        """
        M_star = self.header['M_star']
        R_star = self.header['R_star']
        if self.gyre_version < '6':
            G = pre15140.standard_cgrav
        else:
            G = post15140.standard_cgrav
        return 1.0 / (2 * np.pi) * ((G * M_star / (R_star) ** 3))

    def get_frequencies(self, freq_units):
        """
        Get frequencies in the specicied units. Will use 'Re(omega)' first and 'Re(freq)' otherwise.
        Args:
            freq_units (str): Unit to convert to, must be one of 'uHz', 'mHz', or 'Hz'.
            Re_freq_unit (str, optional): Unit of the 'Re(freq)' column, must be one of 'uHz', 'mHz', or 'Hz'.

        Returns:
            np.rec.array: Frequencies in unit specified by Re_freq_unit.
        """
        unit_dict = {'uHz': 1e6, 'mHz': 1e3, 'Hz': 1e0}
        if 'Re(omega)' in self.header:
            dimless_to_Hz = self._calc_dimless_to_Hz() * unit_dict[freq_units]
            freq_name = 'Re(omega)'
        elif 'Re(freq)' in self.header:  # Assumes freq already in uHz.
            dimless_to_Hz = unit_dict[freq_units] / unit_dict[Re_freq_unit]
            freq_name = 'Re(freq)'
        return self.data[freq_name] * dimless_to_Hz


class GyreProfile:
    def __init__(self, path):
        """
        Gyre profile created by MESA.

        Args:
            path: Path to gyre profile file.
        """
        self.path = path
        if os.path.isfile(path):
            self.fname = os.path.basename(self.path)

        num_zones, mass, radius, luminosity, version = np.loadtxt(f'{path}', max_rows=1)
        num_zones = int(num_zones)
        self.version = int(version)

        self.header = {'num_zones': num_zones, 'star_mass': mass, 'star_radius': radius, 'star_luminosity': luminosity,
                       'version': version}
        if self.version == 100:
            self.columns = ['zone', 'radius', 'mass', 'luminosity', 'pressure', 'temperature', 'density', 'grad_T',
                            'brunt_N2', 'gamma1', 'grad_ad', 'nu_T', 'opacity', 'opacity_partial_T',
                            'opacity_partial_rho', 'total_energy_generation', 'nuclear_energy_generation_partial_T',
                            'nuclear_energy_generation_partial_rho', 'rotation']

            self.formats = [int] + 18 * [float]
        elif self.version == 101:
            self.columns = ['zone', 'radius', 'mass', 'luminosity', 'pressure', 'temperature', 'density', 'grad_T',
                            'brunt_N2', 'gamma1', 'grad_ad', 'nu_T', 'opacity', 'opacity_partial_T',
                            'opacity_partial_rho', 'nuclear_energy_generation', 'nuclear_energy_generation_partial_T',
                            'nuclear_energy_generation_partial_rho', 'rotation']

            self.formats = [int] + 18 * [float]
        elif self.version == 120:
            self.columns = ['zone', 'radius', 'mass', 'luminosity', 'pressure', 'temperature', 'density', 'grad_T',
                            'brunt_N2', 'gamma1', 'grad_ad', 'nu_T', 'opacity', 'opacity_partial_T',
                            'opacity_partial_rho', 'nuclear_energy_generation', 'nuclear_energy_generation_partial_T',
                            'nuclear_energy_generation_partial_rho', 'gravothermal_energy_generation', 'rotation']
            self.formats = [int] + 19 * [float]
        else:
            raise NotImplementedError('Only fileversions 100, 101, and 120 are implemented.')

        self.loaded = False

    def __repr__(self):
        return 'GyreProfile data file at {}'.format(self.path) + '\n' + str(self.header)

    def _load_gyre_profile(self):
        data = np.rec.array(
            np.loadtxt(f'{self.path}', skiprows=1, dtype=list(zip(self.columns, self.formats))))
        return data

    @_LazyProperty
    def data(self):
        data = self._load_gyre_profile()
        self.loaded = True
        return data


def load_profs(hist, prefix='profile', suffix='.data', save_dill=False, mask=None, mask_kwargs=None):
    """
    Load profiles associated with `History` hist.

    Args:
        hist (History):
        prefix (str, optional): Part of profile name before the profile number. Defaults to 'profile'.
        suffix (str, optional): Part of profile name after the profile number. Defaults to '.data'.
        save_dill (bool, optional): If True, will write a `.dill` file containing the `History` data.
        mask (array of bool or function): Mask to apply to column(s).
        mask_kwargs (dict, optional): kwargs to pass to mask if it is a function.

    Returns:
        list of Profile: All profiles in hist.index or those which satisfy mask.
    """
    if hist.index is None:
        return []
    pnums = hist.index[:, 2]

    if mask is not None:
        if hasattr(mask, '__call__'):
            if mask_kwargs is None:
                mask_kwargs = {}
            mask = mask(hist, **mask_kwargs)
        valid_mod = hist.get('model_number')[mask]
        pnums = hist.index[:, 2][np.isin(hist.index[:, 0], valid_mod)]

    profs = []
    for i in pnums:

        try:
            if suffix.endswith('.GYRE'):
                prof = GyreProfile(os.path.join(hist.LOGS, '{}{}{}'.format(prefix, i, suffix)))
                profs.append(prof)
            else:
                prof = Profile(os.path.join(hist.LOGS, '{}{}{}'.format(prefix, i, suffix)), save_dill=save_dill)
                profs.append(prof)
        except IndexError:
            print(os.path.join(hist.LOGS, '{}{}{}'.format(prefix, i, suffix)) +
                  ' not found in {}'.format(hist.index_path))
    return profs


def load_gss(hist, gyre_data_dir='gyre_out', gyre_summary_prefix='profile', gyre_summary_suffix='.data.GYRE.sgyre_l',
             return_pnums=False, use_mask=None, keep_columns='all', gyre_version='7', save_dill=False, reload=False, verbose=False, nanval=-1e99,
                         nanclip=None):
    """
    Load `GyreSummary` associated with `History` hist.

    Args:
        hist:
        gyre_data_dir: Directory containing gyre summary files.
        gyre_summary_prefix (str, optional): Part of gyre summary name before the profile number. Defaults to 'profile'.
        gyre_summary_suffix (str, optional): Part of gyre summary name after the profile number. Defaults to '.data.GYRE.sgyre_l'.
        return_pnums (bool, optional): Defaults to False. If True, will also return profile numbers.
        use_mask (bool, np.array, or function, optional): If True, will exclude pre-main sequence.
                If an array of bools will use that as mask. If a function, will call function(hist) and use that as the mask.
        keep_columns (list of str, optional): Which columns of the history data file to keep.
        gyre_version (str): Gyre version.
        save_dill (bool, optional): If True, will write a `.dill` file containing the `GyreSummary` data.
        reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the gyre summary file.
        verbose (bool, optional): Print extra information.
        nanval (float, optional): Set all values equal to this to NaN.
        nanclip (2 floats, optional): Set all values outside this range to NaN.

    Returns:
        list of GyreSummary or list of list of GyreSummary: If return_pnums is False returns only `GyreSummary`.
            If return_pnums is True also return profile numbers.
    """
    dirpath = os.path.abspath(os.path.join(hist.LOGS, '..', gyre_data_dir))

    use_mask = uf.get_mask(hist, use_mask)

    min_mod, max_mod = hist.get('model_number')[use_mask][[0, -1]]

    pnums = []
    for fname in os.listdir(dirpath):
        fname = os.path.split(fname)[-1]
        if fname.startswith(gyre_summary_prefix) and fname.endswith(gyre_summary_suffix):
            try:
                pnum = int(fname[len(gyre_summary_prefix):-len(gyre_summary_suffix)])
            except ValueError:
                continue
            if pnum in hist.index[:, 2]:  # Only load gss which are in the index
                if min_mod <= hist.index[:, 0][hist.index[:, 2] == pnum] <= max_mod:
                    pnums.append(pnum)
    pnums.sort()

    gss = []
    for pnum in pnums:
        fname = '{}{}{}'.format(gyre_summary_prefix, pnum, gyre_summary_suffix)
        path = os.path.join(dirpath, fname)
        gss.append(GyreSummary(path, keep_columns=keep_columns, gyre_version=gyre_version, save_dill=save_dill, reload=reload,
                     verbose=verbose, nanval=nanval, nanclip=nanclip))

    if return_pnums:
        return list(zip(gss, np.array(pnums)))
    return gss


def load_modes_from_profile(prof, gyre_data_dir='gyre_out', mode_prefix='', mode_suffix='.mgyre', keep_columns='all',
                            gyre_version='7', save_dill=False, reload=False, verbose=False, nanval=-1e99, nanclip=None):
    """
    Load all `GyreMode` associated with `Profile` prof.

    Args:
        prof:
        gyre_data_dir: Directory containing gyre summary files.
        mode_prefix (str, optional): First part of the mode filename.
            If an empty string will use the profile name. Defaults to ''.
        mode_suffix (str, optional): Last part of the gyre mode filename. Defaults to '.mgyre'.
        keep_columns (list of str, optional): Which columns of the mode data file to keep.
        gyre_version (str): Gyre version.
        save_dill (bool, optional): If True, will write a `.dill` file containing the `GyreMode` data.
        reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the mode file.
        verbose (bool, optional): Print extra information.
        nanval (float, optional): Set all values equal to this to NaN.
        nanclip (2 floats, optional): Set all values outside this range to NaN.


    Returns:
        list of GyreMode:
    """
    dirpath = os.path.abspath(os.path.join(prof.LOGS, '..', gyre_data_dir))
    fnames = os.listdir(dirpath)

    if mode_prefix == '':
        mode_prefix = prof.fname

    modes = []
    for fname in fnames:
        if fname.startswith(mode_prefix):
            if fname.endswith(mode_suffix):
                path = os.path.join(dirpath, fname)
                modes.append(GyreMode(path, keep_columns=keep_columns, gyre_version=gyre_version, save_dill=save_dill, reload=reload, verbose=verbose, nanval=nanval, nanclip=nanclip))
    return modes


def load_gs_from_profile(prof, gyre_data_dir='gyre_out', gyre_summary_prefix='', gyre_summary_suffix='.data.GYRE.sgyre_l',
                         keep_columns='all', gyre_version='7', save_dill=False, reload=False, verbose=False, nanval=-1e99,
                         nanclip=None):
    """
    Load the `GyreSummary` associated with `Profile` prof.

    Args:
        prof (Profile):
        gyre_data_dir: Directory containing gyre summary files.
        gyre_summary_prefix (str, optional): First part of gyre summary name. First part of the gyre summary filename.
            If an empty string will use the profile name. Defaults to ''.
        gyre_summary_suffix (str, optional): Part of gyre summary name after the profile number. Defaults to '.data.GYRE.sgyre_l'.
        keep_columns (list of str, optional): Which columns of the history data file to keep.
        gyre_version (str): Gyre version.
        save_dill (bool, optional): If True, will write a `.dill` file containing the `History` data.
        reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the history file.
        verbose (bool, optional): Print extra information.
        nanval (float, optional): Set all values equal to this to NaN.
        nanclip (2 floats, optional): Set all values outside this range to NaN.
    Returns:
        GyreSummary:
    """
    dirpath = os.path.abspath(os.path.join(prof.LOGS, '..', gyre_data_dir))
    fnames = os.listdir(dirpath)

    if gyre_summary_prefix == '':
        gyre_summary_prefix = prof.fname

    for fname in fnames:
        if fname.startswith(gyre_summary_prefix) and fname.endswith(gyre_summary_suffix):
            break
        else:
            fname = ''

    path = os.path.join(dirpath, fname)
    gs = GyreSummary(path, keep_columns=keep_columns, gyre_version=gyre_version, save_dill=save_dill, reload=reload,
                     verbose=verbose, nanval=nanval, nanclip=nanclip)
    return gs


# def load_modes(hist, gyre_summary_dir='gyre_out', gyre_summary_prefix='profile', gyre_summary_suffix='.data.GYRE.sgyre_l',
#                mode_dir='gyre_out/detail', mode_prefix='profile{}.', mode_suffix='.mgyre', keep_columns='all',
#                gyre_version='7', save_dill=False, reload=False, verbose=False, nanval=-1e99, nanclip=None):
#
#     gss, pnums = load_gss(hist, gyre_summary_dir, gyre_summary_prefix, gyre_summary_suffix, return_pnums=True,
#                           keep_columns=keep_columns, gyre_version=gyre_version, save_dill=save_dill, reload=reload,
#                           verbose=verbose, nanval=nanval, nanclip=nanclip)
#
#     dirpath = os.path.abspath(os.path.join(hist.LOGS, '..', mode_dir))
#     fnames = os.listdir(dirpath)
#
#     mode_info = []
#     modes = []
#
#     i = 0
#     for fname in fnames:
#         fname = os.path.split(fname)[-1]
#         prefix_parts = mode_prefix.split('{}')
#         pnum = int(fname[len(prefix_parts[0]):fname.index(prefix_parts[1])])
#         mode_prefix_pnum = mode_prefix.format(pnum)
#         if fname.startswith(mode_prefix_pnum) and fname.endswith(mode_suffix):
#             path = os.path.join(dirpath, fname)
#             md = GyreMode(path, keep_columns=keep_columns, gyre_version=gyre_version, save_dill=save_dill, reload=reload,
#                           verbose=verbose, nanval=nanval, nanclip=nanclip)
#             nu = uf.get_freq(md, kind='mode')
#             l = md.header['l']
#             order_names = ['n_p', 'n_g', 'n_pg']
#             if sum(np.isin(order_names, list(md.header.keys()))) > 1:
#
#                 try:
#                     n_p = md.header['n_p']
#                 except KeyError:
#                     n_p = md.header['n_pg'] + md.header['n_g']
#                     md.header['n_p'] = n_p
#
#                 try:
#                     n_g = md.header['n_g']
#                 except KeyError:
#                     n_g = md.header['n_p'] - md.header['n_pg']
#                     md.header['n_g'] = n_g
#
#                 try:
#                     n_pg = md.header['n_pg']
#                 except KeyError:
#                     n_pg = md.header['n_p'] - md.header['n_g']
#                     md.header['n_pg'] = n_pg
#
#             else:
#                 raise ValueError('Not enough information in header to determine n_p, n_g, and n_pg.')
#
#             md.header['profile_number'] = pnum
#
#             for key in ['n_p', 'n_g', 'n_pg']:
#                 if key in md.header and key in locals().keys():
#                     if md.header[key] != locals()[key]:
#                         raise ValueError(f'Mismatch between mode {key} and calculated {key}!')
#
#                 else:
#                     md.header[key] = locals()[key]
#
#             mode_info.append([pnum, l, nu, n_pg, n_p, n_g, i])
#             modes.append(md)
#             i += 1
#
#     mode_info = np.rec.array(mode_info, names=['pnum', 'l', 'nu', 'n_pg', 'n_p', 'n_g', 'i'])
#     mode_info = np.sort(mode_info, order=['pnum', 'l', 'nu', 'n_pg'])
#     modes = np.asarray([modes[i] for i in mode_info.i], dtype=object)
#
#     mode_info.i = np.arange(len(mode_info))
#
#     return modes


def naive_merge_hists(base_hist, hists):
    """
    Merge two `History` objects. This function simply stacks the history data onto a new copy of base_hist.

    Args:
        base_hist (History):
        hists (list of History): Histories to stack.

    Returns:
        History
    """
    new_hist = copy.copy(base_hist)
    new_hist.data = np.lib.recfunctions.stack_arrays([h.data for h in hists], asrecarray=True, usemask=False)
    return new_hist


def load_gss_to_hist(hist, gyre_data_dir='gyre_out', gyre_summary_prefix='profile',
                     gyre_summary_suffix='.data.GYRE.sgyre_l', use_mask=None, keep_columns='all',
                     gyre_version='7', save_dill=False, reload=False, verbose=False, nanval=-1e99, nanclip=None):
    """
    Load ``GyreSummary`` and profile numbers associated with ``History`` hist and place in the attribute ``History.gsspnum``.
    This is equivalent to doing ``hist.gsspnum = load_gss(..., return_pnums=True, ...)``.

    Args:
        hist:
        gyre_data_dir: Directory containing gyre summary files.
        gyre_summary_prefix (str, optional): Part of gyre summary name before the profile number. Defaults to 'profile'.
        gyre_summary_suffix (str, optional): Part of gyre summary name after the profile number. Defaults to '.data.GYRE.sgyre_l'.
        use_mask (bool, np.array, or function, optional): If True, will exclude pre-main sequence.
                If an array of bools will use that as mask. If a function, will call function(hist) and use that as the mask.
        keep_columns (list of str, optional): Which columns of the history data file to keep.
        gyre_version (str): Gyre version.
        save_dill (bool, optional): If True, will write a `.dill` file containing the `GyreSummary` data.
        reload (bool, optional): If True, will ignore a pre-existing `.dill` file and reload from the gyre summary file.
        verbose (bool, optional): Print extra information.
        nanval (float, optional): Set all values equal to this to NaN.
        nanclip (2 floats, optional): Set all values outside this range to NaN.

    Returns:
        list of GyreSummary or list of list of GyreSummary: If return_pnums is False returns only `GyreSummary`.
            If return_pnums is True also return profile numbers.
    """
    hist.gsspnum = load_gss(hist, gyre_data_dir=gyre_data_dir, gyre_summary_prefix=gyre_summary_prefix,
                            gyre_summary_suffix=gyre_summary_suffix, return_pnums=True, use_mask=use_mask,
                            keep_columns=keep_columns, gyre_version=gyre_version, save_dill=save_dill, reload=reload,
                            verbose=verbose, nanval=nanval, nanclip=nanclip)
    return hist



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Numerical constants used by MESA in ``$MESA_DIR/const/public/const_def.f90``. This file is generated by ``wsssss/dev/create_const_files.py``.
"""

version = 'r24.03.1'
"""MESA version used to generate this file."""

max_extra_inlists = 5
"""number of inlists an inlist can depend on"""

pi = 3.1415926535897932384626433832795028841971693993751E0
"""pi = 3.1415926535897932384626433832795028841971693993751E0"""

pi2 = pi * pi
"""pi2 = pi * pi"""

pi4 = 4*pi
"""pi4 = 4*pi"""

eulercon = 0.577215664901532861E0
"""eulercon = 0.577215664901532861E0"""

eulernum = 2.71828182845904523536028747135266249E0
"""eulernum = 2.71828182845904523536028747135266249E0"""

ln2 = 6.9314718055994529E-01
"""log(2d0)"""

ln3 = 1.0986122886681096E+00
"""log(3d0)"""

lnpi = 1.14472988584940017414343
"""log(pi)"""

ln10 = 2.3025850929940455
"""log(10d0)"""

iln10 = 0.43429448190325187
"""1d0/log(10d0)"""

a2rad = pi/180.0E0
"""angle to radians"""

rad2a = 180.0E0/pi
"""radians to angle"""

one_third = 1E0/3E0
"""one_third = 1E0/3E0"""

two_thirds = 2E0/3E0
"""two_thirds = 2E0/3E0"""

four_thirds = 4E0/3E0
"""four_thirds = 4E0/3E0"""

five_thirds = 5E0/3E0
"""five_thirds = 5E0/3E0"""

one_sixth = 1E0/6E0
"""one_sixth = 1E0/6E0"""

four_thirds_pi = four_thirds*pi
"""four_thirds_pi = four_thirds*pi"""

ln4pi3 = 1.4324119583011810E0
"""log(4*pi/3)"""

two_13 = 1.2599210498948730E0
"""pow(2d0,1d0/3d0)"""

four_13 = 1.5874010519681994E0
"""pow(4d0,1d0/3d0)"""

sqrt2 = 1.414213562373095E0
"""sqrt(2)"""

sqrt_2_div_3 = 0.816496580927726E0
"""sqrt(2/3)"""

avo = 6.02214076E23
"""Avogadro constant (mole^-1)"""

amu = 1E0 / avo
"""atomic mass unit (g)"""

clight = 2.99792458E10
"""speed of light in vacuum (cm s^-1)"""

qe = (clight/10E0) * 1.602176634E-19
"""elementary charge (esu # (g cm^3 s^-2)^(1/2))"""

kerg = 1.380649E-16
"""kerg = 1.380649E-16"""

boltzm = kerg
"""Boltzmann constant (erg K^-1)"""

planck_h = 6.62607015E-27
"""Planck constant (erg s)"""

hbar = planck_h / (2*pi)
"""hbar = planck_h / (2*pi)"""

cgas = boltzm*avo
"""ideal gas constant (erg K^-1)"""

ev2erg = 1.602176634E-12
"""electron volt (erg)"""

mev_to_ergs = 1E6*ev2erg
"""mev_to_ergs = 1E6*ev2erg"""

mev_amu = mev_to_ergs/amu
"""mev_amu = mev_to_ergs/amu"""

mev2gr = 1E6*ev2erg/(clight*clight)
"""MeV to grams"""

qconv = mev_to_ergs*avo
"""qconv = mev_to_ergs*avo"""

kev = kerg / ev2erg
"""converts temp to ev (ev K^-1)"""

boltz_sigma = (pi*pi * boltzm*boltzm*boltzm*boltzm) / (60 * hbar*hbar*hbar * clight*clight)
"""Stefan-Boltzmann constant (erg cm^-2 K^-4 s^-1)"""

crad = boltz_sigma*4/clight
"""radiation density constant, AKA 'a' (erg cm^-3 K^-4)# Prad #  crad * T^4 / 3"""

au = 1.49597870700E13
"""exact value defined by IAU 2009, 2012 (cm)"""

pc = (3.600E3 * rad2a) * au
"""parsec, by definition (cm)"""

dayyer = 365.25E0
"""days per (Julian) year"""

secday = 24*60*60
"""seconds in a day"""

secyer = secday*dayyer
"""seconds per year"""

ly = clight*secyer
"""light year (cm)"""

mn = 1.67492749804E-24
"""neutron mass (g)"""

mp = 1.67262192369E-24
"""proton mass (g)"""

me = 9.1093837015E-28
"""electron mass (g)"""

rbohr = 5.29177210903E-9
"""Bohr radius (cm)"""

fine = 7.2973525693E-3
"""fine-structure constant"""

hion = 13.605693122994E0
"""Rydberg constant (eV)"""

sige = 6.6524587321E-25
"""Thomson cross section (cm^2)"""

weinberg_theta  = 0.22290E0
"""sin**2(theta_weinberg)"""

num_neu_fam = 3.0E0
"""number of neutrino flavors #  3.02 plus/minus 0.005 (1998)"""

standard_cgrav = 6.67430E-8
"""gravitational constant (g^-1 cm^3 s^-2)"""

mu_sun = 1.3271244E26
"""mu_sun = 1.3271244E26"""

mu_earth = 3.986004E20
"""mu_earth = 3.986004E20"""

mu_jupiter = 1.2668653E23
"""mu_jupiter = 1.2668653E23"""

agesun = 4.57E9
"""solar age (years) from Bahcall et al, ApJ 618 (2005) 1049-1056."""

msun = mu_sun / standard_cgrav
"""solar mass (g)# gravitational mass, not baryonic"""

rsun = 6.957E10
"""solar radius (cm), IAU 2015 Resolution B3"""

lsun = 3.828E33
"""solar luminosity (erg s^-1), IAU 2015 Resolution B3"""

teffsun = 5772.0E0
"""solar effective temperature (K), IAU 2015 Resolution B3"""

loggsun = 4.4380676273031332
"""log10(mu_sun/(Rsun*Rsun)), can't call log10 because we don't have math_lib at this point"""

mbolsun = 4.74E0
"""Bolometric magnitude of the Sun, IAU 2015 Resolution B2"""

m_earth = mu_earth/standard_cgrav
"""earth mass (g)"""

r_earth = 6.3781E8
"""earth equatorial radius (cm)"""

r_earth_polar = 6.3568E8
"""earth polar radius (cm)"""

m_jupiter = mu_jupiter/standard_cgrav
"""jupiter mass (g)"""

r_jupiter = 7.1492E9
"""jupiter equatorial radius (cm)"""

r_jupiter_polar = 6.6854E9
"""jupiter polar radius (cm)"""

semimajor_axis_jupiter = 7.7857E13
"""jupiter semimajor axis (cm)"""

arg_not_provided = -9E99
"""arg_not_provided = -9E99"""

missing_value = arg_not_provided
"""missing_value = arg_not_provided"""

crystallized = -1
"""crystallized = -1"""

no_mixing = 0
"""no_mixing = 0"""

convective_mixing = 1
"""convective_mixing = 1"""

overshoot_mixing = 2
"""overshoot_mixing = 2"""

semiconvective_mixing = 3
"""semiconvective_mixing = 3"""

thermohaline_mixing = 4
"""thermohaline_mixing = 4"""

rotation_mixing = 5
"""rotation_mixing = 5"""

rayleigh_taylor_mixing = 6
"""rayleigh_taylor_mixing = 6"""

minimum_mixing = 7
"""minimum_mixing = 7"""

anonymous_mixing = 8
"""AKA 'WTF_mixing'"""

leftover_convective_mixing = 9
"""for regions with non-zero conv_vel that are not unstable to convection  used for time dependent convection"""

phase_separation_mixing = 10
"""phase_separation_mixing = 10"""

number_of_mixing_types =  phase_separation_mixing+1
"""number_of_mixing_types =  phase_separation_mixing+1"""

